from lyzr.base.llm import LyzrLLMFactory, LiteLLM
from lyzr.base.llms import LLM, get_model
from lyzr.base.service import LyzrService
from lyzr.base.vector_store import LyzrVectorStoreIndex
from lyzr.base.retrievers import LyzrRetriever
from lyzr.base.prompt import LyzrPromptFactory
from lyzr.base.base import ChatMessage, UserMessage, SystemMessage, AssistantMessage

__all__ = [
    "LyzrLLMFactory",
    "LyzrService",
    "LyzrVectorStoreIndex",
    "LLM",
    "get_model",
    "LyzrRetriever",
    "LiteLLM",
    "LyzrPromptFactory",
    "ChatMessage",
    "UserMessage",
    "SystemMessage",
    "AssistantMessage",
]



"""
Base classes and configuration models for the DataAnalyzr class.
"""

# standart-library imports
import json
import logging
import warnings
from enum import Enum
from typing import Union, Any

# third-party imports
from aenum import Enum as AEnum
from pydantic import BaseModel, Field, AliasChoices, ConfigDict, model_validator

# local imports
from lyzr.base import LiteLLM

warnings.filterwarnings("ignore", category=UserWarning)


class AnalysisTypes(str, AEnum):
    """A mutable enumeration of analysis types."""

    sql = "sql"
    ml = "ml"
    skip = "skip"


class OutputTypes(str, Enum):
    """An enumeration of output types."""

    visualisation = "visualisation"
    visualisations = "visualisation"
    vizualisation = "visualisation"
    vizualisations = "visualisation"
    visualization = "visualisation"
    visualizations = "visualisation"
    image = "visualisation"
    images = "visualisation"
    insights = "insights"
    insight = "insights"
    recommendations = "recommendations"
    recommendation = "recommendations"
    tasks = "tasks"
    task = "tasks"


class ParamsDict(BaseModel):
    """Configuration model for `params` attribute of factory classes and the DataAnalyzr class."""

    max_retries: Union[int, None] = Field(default=None)
    time_limit: Union[int, None] = Field(default=None)
    auto_train: Union[bool, None] = Field(default=True)


class ContextDict(BaseModel):
    """
    A class to manage and validate context data for analysis, visualisation, insights, recommendations, and tasks.

    Attributes:
        analysis (Union[str, None]): The analysis context, with alias choices for validation.
        visualisation (Union[str, None]): The visualisation context, with alias choices for validation.
        insights (Union[str, None]): The insights context, with alias choices for validation.
        recommendations (Union[str, None]): The recommendations context, with alias choices for validation.
        tasks (Union[str, None]): The tasks context, with alias choices for validation.
        model_config (ConfigDict): Configuration for model validation, ensuring assignment validation.

    Methods:
        validate_context(cls, data):
            Class method to validate and format context data before model instantiation.
        validate_from_string(self, context_str: str):
            Validates and formats context data from a string.
        validate_from_dict(self, context_dict: dict):
            Validates and formats context data from a dictionary.
        validate(self, context=None, **kwargs):
            Validates and formats context data from either a string, dictionary, or keyword arguments.
        reset_values(self):
            Resets all context fields to empty strings.
    """

    analysis: Union[str, None] = Field(
        default_factory=str, validation_alias=AliasChoices("analysis", "analyses")
    )
    visualisation: Union[str, None] = Field(
        default_factory=str,
        validation_alias=AliasChoices(
            "visualisation",
            "visualisations",
            "visualization",
            "visualizations",
            "image",
        ),
    )
    insights: Union[str, None] = Field(
        default_factory=str, validation_alias=AliasChoices("insights", "insight")
    )
    recommendations: Union[str, None] = Field(
        default_factory=str,
        validation_alias=AliasChoices("recommendations", "recommendation"),
    )
    tasks: Union[str, None] = Field(
        default_factory=str, validation_alias=AliasChoices("tasks", "task")
    )
    model_config = ConfigDict(validate_assignment=True)

    @model_validator(mode="before")
    @classmethod
    def validate_context(cls, data):
        if isinstance(data, str):
            data = data.strip() + "\n\n" if data.strip() != "" else ""
            return {fname: data for fname in cls.model_fields}
        if not isinstance(data, dict):
            return data
        for key, value in data.items():
            if isinstance(value, str) and value.strip() != "":
                data[key] = value.strip() + "\n\n"
        return data

    def validate_from_string(self, context_str: str):
        if not isinstance(context_str, str):
            return self
        context_str = context_str.strip()
        if context_str != "":
            context_str = context_str + "\n\n"
        for key, value in self:
            if value is None or not isinstance(value, str) or value.strip() == "":
                setattr(self, key, context_str)
            else:
                setattr(self, key, value.strip() + "\n\n")
        return self

    def validate_from_dict(self, context_dict: dict):
        if isinstance(context_dict, dict):
            context_dict = ContextDict.model_validate_json(json.dumps(context_dict))
        if not isinstance(context_dict, ContextDict):
            return self
        for key, value in context_dict:
            if isinstance(value, str) and value.strip() != "":
                setattr(self, key, value.strip() + "\n\n")
        return self

    def validate(self, context=None, **kwargs):
        if isinstance(context, str):
            self = self.validate_from_string(context)
        if isinstance(context, (dict, ContextDict)):
            self = self.validate_from_dict(context)
        return self.validate_from_dict(kwargs)

    def reset_values(self):
        for key, _ in self:
            setattr(self, key, "")
        return self


class VectorStoreConfig(BaseModel):
    """Configuration model for vector store initialisation settings."""

    path: Union[str, None] = Field(
        default=None,
        validation_alias=AliasChoices("vector_store_path", "vstore_path", "path"),
    )
    remake_store: Union[bool, None] = Field(
        default=False, validation_alias=AliasChoices("remake", "remake_store")
    )


class FactoryBaseClass:
    """
    A base class for creating factory objects that interact with a language model, logging system, and vector store.

    Attributes:
        llm (LiteLLM): The language model instance.
        context (str): The context or environment in which the factory operates.
        logger (logging.Logger): The logger instance for logging messages.
        vector_store (ChromaDBVectorStore): The vector store for managing and retrieving vectors.
        output (Any): The output generated by the factory.
        code (str): The code to be executed or processed.
        params (ParamsDict): The parameters dictionary containing configuration settings.

    Methods:
        __init__(llm: LiteLLM, logger: logging.Logger, context: str, vector_store: ChromaDBVectorStore,
            max_retries: int, time_limit: int, auto_train: bool, llm_kwargs: dict) -> None:
            Initializes the factory with the given parameters and sets up the language model and vector store.

        generate_output(**kwargs):
            Abstract method to generate output. Must be implemented by subclasses.

        get_prompt_messages(**kwargs) -> list:
            Abstract method to get prompt messages. Must be implemented by subclasses.

        extract_and_execute_code(llm_response: str):
            Abstract method to extract and execute code from the language model response. Must be implemented by subclasses.

        auto_train(user_input: str, code: str, **kwargs) -> None:
            Abstract method to perform auto-training. Must be implemented by subclasses.
    """

    from lyzr.data_analyzr.vector_store_utils import ChromaDBVectorStore

    llm: LiteLLM
    context: str
    logger: logging.Logger
    vector_store: ChromaDBVectorStore
    output: Any
    code: str
    params: ParamsDict

    def __init__(
        self,
        llm: LiteLLM,
        logger: logging.Logger,
        context: str,
        vector_store: ChromaDBVectorStore,
        max_retries: int,
        time_limit: int,
        auto_train: bool,
        llm_kwargs: dict,
    ) -> None:
        self.llm = llm
        model_kwargs = dict(seed=123, temperature=0.1, top_p=0.5)
        model_kwargs.update(llm_kwargs)
        self.llm.set_model_kwargs(model_kwargs=model_kwargs)
        self.context = context
        self.logger = logger
        self.vector_store = vector_store
        self.params = ParamsDict(
            max_retries=max_retries,
            time_limit=time_limit,
            auto_train=True if auto_train is None else auto_train,
        )
        self.code = None
        self.output = None

    def generate_output(self, **kwargs):
        raise NotImplementedError

    def get_prompt_messages(self, **kwargs) -> list:
        raise NotImplementedError

    def extract_and_execute_code(self, llm_response: str):
        raise NotImplementedError

    def auto_train(self, user_input: str, code: str, **kwargs) -> None:
        raise NotImplementedError



from lyzr.chat.chatbot import ChatBot
from lyzr.base.llm import LyzrLLMFactory
from lyzr.qa.qa_bot import QABot
from lyzr.base.service import LyzrService
from lyzr.base.vector_store import LyzrVectorStoreIndex
from lyzr.formula_generator import FormulaGen
from lyzr.data_analyzr import DataAnalyzr
from lyzr.voicebot import VoiceBot
from lyzr.qa.search_agent import SearchAgent
from lyzr.summarizer import Summarizer
from lyzr.generator import Generator

__all__ = [
    "LyzrLLMFactory",
    "LyzrService",
    "LyzrVectorStoreIndex",
    "QABot",
    "ChatBot",
    "FormulaGen",
    "DataAnalyzr",
    "VoiceBot",
    "SearchAgent",
    "Summarizer",
    "Generator",
]



"""
TxttoSQLFactory class - generate and execute SQL queries from text inputs.
"""

# standard library imports
import re
import logging
import traceback
from typing import Union

# third-party imports
import pandas as pd

# local imports
from lyzr.base.llm import LiteLLM
from lyzr.base.prompt import LyzrPromptFactory
from lyzr.base.errors import MissingValueError
from lyzr.data_analyzr.models import FactoryBaseClass
from lyzr.data_analyzr.db_connector import DatabaseConnector
from lyzr.base.base import ChatMessage, UserMessage, SystemMessage
from lyzr.data_analyzr.vector_store_utils import ChromaDBVectorStore
from lyzr.data_analyzr.analysis_handler.utils import (
    extract_sql,
    iterate_llm_calls,
    handle_analysis_output,
)


class TxttoSQLFactory(FactoryBaseClass):
    """
    TxttoSQLFactory is a class that facilitates the conversion of natural language text into SQL queries using a language model.

    This class integrates with a database connector to execute the generated SQL queries and provides logging and context
    management functionalities. It also supports automatic training data generation and management through a vector store.

    Attributes:
        connector (DatabaseConnector): The database connector for executing generated SQL queries.

    Methods:
        generate_output(user_input: str, **kwargs) -> Union[str, pd.DataFrame, dict[str, pd.DataFrame], None]:
            Run an analysis on the provided user input by generating and executing SQL queries.

        get_prompt_messages(user_input: str) -> list[ChatMessage]:
            Generates a list of chat messages based on the user's input.

        _get_sql_prompt(user_input: str, question_sql_list: list, ddl_list: list, doc_list: list) -> list[ChatMessage]:
            Generates a list of ChatMessages based on the provided user input and context.

        extract_and_execute_code(llm_response: str):
            Extracts an SQL query from the given LLM response and executes it.

        code_cleaner(code) -> str:
            Handler for cleaning the extracted code before execution.

        _handle_create_table_sql(sql_query: str):
            Handles the execution of a SQL query when table creation is involved.

        auto_train(user_input: str, code: str, **kwargs):
            Adds the user input and generated SQL to the vector store if the auto_train flag is set.

        _generate_question(sql: str, **kwargs) -> str:
            Generates a question based on the provided SQL query.
    """

    def __init__(
        self,
        llm: LiteLLM,
        db_connector: DatabaseConnector,
        logger: logging.Logger,
        context: str,
        vector_store: ChromaDBVectorStore,
        max_retries: int = None,
        time_limit: int = None,
        auto_train: bool = None,
        **llm_kwargs,
    ):
        """
        Initialize a TxttoSQLFactory instance.

        Args:
            llm (LiteLLM): The llm instance to be used.
            db_connector (DatabaseConnector): The database connector for executing generated SQL queries.
            logger (logging.Logger): Logger for logging events and errors.
            context (str): The context for the given query. If empty, pass "".
            vector_store (ChromaDBVectorStore): The vector store for managing related queries and database documentation.
            max_retries (int, optional): Maximum number of retries for analysis operations. Defaults to 10 if None.
            time_limit (int, optional): Time limit for analysis in seconds. Defaults to 45 if None.
            auto_train (bool, optional): Whether to automatically add to training data. Defaults to True.
            **llm_kwargs: Additional keyword arguments for the language model.

        Example Usage:
            from lyzr.base.llm import LiteLLM
            from lyzr.data_analyzr.vector_store_utils import ChromaDBVectorStore
            from lyzr.data_analyzr.analysis_handler import TxttoSQLFactory

            llm = LiteLLM.from_defaults(model="gpt-4o")
            logger = logging.getLogger(__name__)
            context = "Context for the analysis."
            db_connector = DatabaseConnector(...)
            vector_store = ChromaDBVectorStore(path="path/to/vector_store", logger=logger)
            analysis_factory = TxttoSQLFactory(
                llm=llm,
                db_connector=db_connector,
                logger=logger,
                context=context,
                vector_store=vector_store
            )
        """
        super().__init__(
            llm=llm,
            logger=logger,
            context=context,
            vector_store=vector_store,
            max_retries=10 if max_retries is None else max_retries,
            time_limit=45 if time_limit is None else time_limit,
            auto_train=auto_train,
            llm_kwargs=llm_kwargs,
        )
        self.connector = db_connector

    def generate_output(
        self,
        user_input: str,
        **kwargs,  # max_tries=3, time_limit=30, auto_train=True
    ) -> Union[str, pd.DataFrame, dict[str, pd.DataFrame], None]:
        """
        Run an analysis on the provided user input by generating and executing SQL queries.

        This is the primary method for generating and executing SQL analysis.
            - Processes the user input to generate SQL analysis output.
            - Handles retries and time limits for the SQL extraction and execution process.
            - Uses iterate_llm_calls to handle retries and time limits.

        Args:
            user_input (str): The user's question input for generating SQL queries. Should be in natural language.

        Keyword Args:
            max_retries (int, optional): Maximum number of retries for generating SQL queries. Defaults to self.params.max_retries.
            time_limit (int, optional): Time limit for the analysis process in seconds. Defaults to self.params.time_limit.
            auto_train (bool, optional): Whether to automatically add the analysis to the training data. Defaults to self.params.auto_train.

        Raises:
            MissingValueError: If the user input is empty or None.

        Returns:
            Union[str, pd.DataFrame, dict[str, pd.DataFrame], None]: The result of the SQL analysis.

        Example usage:
            output = analysis_factory.generate_output("Your question here.")
            print(output)
        """
        user_input = user_input.strip()
        if user_input is None or user_input == "":
            raise MissingValueError("A user input is required for analysis.")
        messages = self.get_prompt_messages(user_input)
        self.extract_and_execute_code = iterate_llm_calls(
            max_retries=kwargs.pop("max_retries", self.params.max_retries),
            llm=self.llm,
            llm_messages=messages,
            logger=self.logger,
            log_messages={
                "start": f"Starting SQL analysis for query: {user_input}",
                "end": f"Ending SQL analysis for query: {user_input}",
            },
            time_limit=kwargs.pop("time_limit", self.params.time_limit),
        )(self.extract_and_execute_code)
        self.output = handle_analysis_output(self.extract_and_execute_code())
        self.auto_train(user_input, self.code, **kwargs)
        return self.output

    def get_prompt_messages(self, user_input: str) -> list[ChatMessage]:
        """
        Generates a list of chat messages based on the user's input.

        Retrieve similar SQL questions, related DDL statements, and relevant documentation
        from the vector store based on the user's input. The list of messages includes system
        messages with context, DDL statements, documentation, and historical examples, followed
        by the user's input.

        Args:
            user_input (str): The input provided by the user.

        Returns:
            list[ChatMessage]: A list of chat messages generated based on the user's input.
        """
        question_sql_list = self.vector_store.get_related_sql_queries(user_input)
        ddl_list = self.vector_store.get_related_ddl(user_input)
        doc_list = self.vector_store.get_related_documentation(user_input)
        messages = self._get_sql_prompt(
            user_input=user_input,
            question_sql_list=question_sql_list,
            ddl_list=ddl_list,
            doc_list=doc_list,
        )
        return messages

    def _get_sql_prompt(
        self,
        user_input: str,
        question_sql_list: list,
        ddl_list: list,
        doc_list: list,
    ) -> list[ChatMessage]:
        """
        Generates a list of ChatMessages based on the provided user input and context.

        Args:
            user_input (str): The user's input or query that needs to be converted to SQL.
            question_sql_list (list): A list of dictionaries containing example questions and their corresponding SQL queries.
            ddl_list (list): A list of Data Definition Language (DDL) statements relevant to the context.
            doc_list (list): A list of additional documentation or context that might help in generating the SQL query.

        Returns:
            list[ChatMessage]: A list of ChatMessage objects that form the complete prompt for generating the SQL query.
        """
        system_message_sections = ["context", "external_context", "task"]
        system_message_dict = {"context": self.context}

        if len(ddl_list) > 0:
            system_message_sections.append("ddl_addition_text")
            system_message_dict["ddl"] = ""
            for ddl_item in ddl_list:
                system_message_dict["ddl"] += f"{ddl_item}\n"

        if len(doc_list) > 0:
            system_message_sections.append("doc_addition_text")
            system_message_dict["doc"] = ""
            for doc_item in doc_list:
                system_message_dict["doc"] += f"{doc_item}\n"
        if len(question_sql_list) > 0:
            system_message_sections.append("history")
        messages = [
            LyzrPromptFactory(
                name="txt_to_sql",
                prompt_type="system",
                use_sections=system_message_sections,
            ).get_message(**system_message_dict)
        ]

        for example in question_sql_list:
            if example is not None:
                if "question" in example and "sql" in example:
                    messages.append(UserMessage(content=example["question"]))
                    messages.append(SystemMessage(content=example["sql"]))

        messages.append(UserMessage(content=user_input))
        return messages

    def extract_and_execute_code(self, llm_response: str):
        """
        Extracts an SQL query from the given LLM response and executes it.
        To be used as a callback function for iterate_llm_calls.

        Args:
            llm_response (str): The response from the language model containing the SQL query.

        Returns:
            Any: The result of the executed SQL query.

        Logs:
            - Info: Extracted SQL query.

        Procedure:
            - Extracts the SQL query from the LLM response.
            - Checks if the SQL query is a CREATE TABLE statement.
            - Executes the SQL query using the database connector.
            - If the SQL query is a CREATE TABLE statement, passes control to _handle_create_table_sql.
        """
        sql_query, analysis_output = None, None
        sql_query = self.code_cleaner(extract_sql(llm_response))
        self.logger.info(f"Extracted SQL query:\n{sql_query}")
        match = re.search("CREATE TABLE", sql_query, re.IGNORECASE)
        if match is not None:
            analysis_output = self._handle_create_table_sql(sql_query)
        else:
            analysis_output = self.connector.run_sql(sql_query)
        self.code = sql_query
        return analysis_output

    def code_cleaner(self, code) -> str:
        """Handler for cleaning the extracted code before execution."""
        return code

    def _handle_create_table_sql(self, sql_query: str):
        """
        Handles the execution of a SQL query when table creation is involved.

        Tries to execute the SQL query without the CREATE TABLE part first.
        If that fails, it retries the original SQL query.

        Args:
            sql_query (str): The SQL query to be executed.

        Returns:
            Any: The result of the executed SQL query.

        Raises:
            Logs an error and attempts to re-run the original SQL query if an exception occurs during execution.

        Logs:
            - Error: When an exception occurs during SQL execution.
        """
        match = re.search("SELECT", sql_query, re.IGNORECASE)
        if match is not None:
            sub_query = sql_query[match.start() :]
        else:
            sub_query = sql_query
        try:
            return self.connector.run_sql(sub_query)
        except Exception as e:
            self.logger.error(
                f"Error running SQL sub-query.{e.__class__.__name__}: {e}",
                extra={
                    "input_kwargs": {
                        "sql_query": sql_query,
                        "sub_query": sub_query,
                    },
                    "traceback": traceback.format_exc().splitlines(),
                },
            )
            return self.connector.run_sql(sql_query)

    def auto_train(self, user_input: str, code: str, **kwargs):
        """
        Adds the user input and generated SQL to the vector store if the auto_train flag is set.

        Args:
            user_input (str): The user's input query.
            code (str): The SQL query generated from the user input.
            auto_train (bool): Whether to automatically add the analysis to the training data,
                defaults to self.params.auto_train.

        Returns:
            Any: The output generated by the model, if available.

        Procedure:
            - Checks if the auto_train flag is set and the output is not None.
            - Generates a question based on the SQL query if the user input is empty or None.
            - Adds the user input and SQL query to the training data in the vector store.
        """
        if (
            kwargs.pop("auto_train", self.params.auto_train)
            and self.output is not None
            and len(self.output) > 0
        ):
            if user_input is None or user_input.strip() == "":
                user_input = self._generate_question(code)
            if code is not None and code.strip() != "":
                self.vector_store.add_training_data(question=user_input, sql=code)
        return self.output

    def _generate_question(self, sql: str, **kwargs) -> str:
        """
        Generates a question based on the provided SQL query.
        To be used when the user input is empty or None.

        Args:
            sql (str): The SQL query for which to generate a question.
            **kwargs: Additional keyword arguments to pass to the llm.

        Returns:
            str: The generated question based on the SQL query.
        """
        kwargs["max_tokens"] = 500
        output = self.llm.run(
            messages=[
                LyzrPromptFactory(
                    name="sql_question_gen", prompt_type="system"
                ).get_message(),
                UserMessage(content=sql),
            ],
            **kwargs,
        )
        return output.message.content



"""
Utility functions for the data analyzr module in Lyzr.
"""

# standart-library imports
import re
import time
import string
import logging
import hashlib
import traceback
from typing import Union
from functools import wraps

# third-party imports
import numpy as np
import pandas as pd


def deterministic_uuid(content: Union[str, bytes, list] = None):
    """
    Generate a deterministic UUID based on the provided content.

    Creates a deterministic UUID by hashing the provided content using MD5.
    The content can be a string, bytes, or a list. If no content is provided, the current time is used.

    Args:
        content (Union[str, bytes, list], optional):
            The content to be hashed. Can be a string, bytes, or a list. Defaults to None.

    Returns:
        str: The generated deterministic UUID as a hexadecimal string.
    """
    if content is None:
        content = str(time.time())
    if isinstance(content, list):
        content = "/".join([str(x) for x in content if x is not None])
    if isinstance(content, str):
        content = content.encode("utf-8")
    hash_object = hashlib.md5(content)
    return hash_object.hexdigest()


def translate_string_name(name: str) -> str:
    """
    Converts a given string into a standardized format.

    Args:
        name (str): The input string to be transformed.

    Returns:
        str: The transformed string.

    Procedure:
    - Converts all characters to lowercase.
    - Strips leading and trailing whitespace.
    - Replaces all punctuation and spaces with underscores.
    - Removes leading and trailing underscores.
    """
    punc = string.punctuation + " "
    new_name = (
        name.lower().strip().translate(str.maketrans(punc, "_" * len(punc))).strip("_")
    )
    return re.sub(r"_+", "_", new_name)


def format_analysis_output(output_df, name: str = None) -> str:
    """
    Formats the details of a DataFrame, Series, list of DataFrames, or dictionary of DataFrames into a string.

    Formats inputs them into a string representation, depending on the input type:
    - pandas Series: Converts the Series to a DataFrame and formats it.
    - list: Recursively formats each element in the list.
    - dictionary: Recursively formats each value in the dictionary.
    - Any other input: Converts the input to a string.
    - pandas DataFrame: Uses a helper function `df_details_with_describe` to format the DataFrame details.

    Args:
        output_df (Union[pd.DataFrame, pd.Series, list, dict, Any]): The input data to be formatted.
        name (str, optional): An optional name to be used in the formatting. Defaults to None.

    Returns:
        str: The formatted string representation of the input data.
    """
    if isinstance(output_df, pd.Series):
        output_df = output_df.to_frame()
    if isinstance(output_df, list):
        return "\n".join([format_analysis_output(df) for df in output_df])
    if isinstance(output_df, dict):
        return "\n".join(
            [format_analysis_output(df, name) for name, df in output_df.items()]
        )
    if not isinstance(output_df, pd.DataFrame):
        return str(output_df)
    else:
        return df_details_with_describe(output_df, name)


def df_details_with_describe(
    output_df: Union[None, pd.DataFrame], name: str = None
) -> str:
    """
    Provides a detailed string representation of a DataFrame, including a snapshot and descriptive statistics.

    Generates a string that includes:
    - The name of the DataFrame.
    - A snapshot of the DataFrame (first and last 50 rows if the DataFrame is large).
    - Descriptive statistics of the DataFrame columns if the DataFrame is large.
    - If input is None, the function returns a string indicating that the DataFrame is None.

    Args:
        output_df (Union[None, pd.DataFrame]): The DataFrame to be described.
        name (str, optional): The name of the DataFrame. Defaults to "Dataframe".

    Returns:
        str: A string representation of the DataFrame details.
    """
    name = name or "Dataframe"
    if output_df is None:
        return f"{name}: None"
    if output_df.size > 100:
        df_display = pd.concat([output_df.head(50), output_df.tail(50)], axis=0)
        df_string = f"DataFrame name: {name}\nDataFrame snapshot:\n{_df_to_string(df_display)}\n\nDataFrame column details:\n{_df_to_string(output_df.describe())}"
    else:
        df_string = f"{name}:\n{_df_to_string(output_df)}"
    return df_string


def _df_to_string(output_df: pd.DataFrame) -> str:
    """
    Converts a DataFrame to a formatted string representation.

    Args:
        output_df (pd.DataFrame): The DataFrame to be converted to a string.

    Returns:
        str: The formatted string representation of the DataFrame.

    Procedure:
    - Convert all column names to strings.
    - Identify and convert columns with date or time information to datetime objects.
    - Format the DataFrame to a string with specific formatting for floats and datetime columns.
    """
    output_df.columns = [str(col) for col in output_df.columns.tolist()]
    datetimecols = [
        col
        for col in output_df.columns.tolist()
        if ("date" in col.lower() or "time" in col.lower())
        and isinstance(output_df[col].dtype, np.number)
    ]
    if "timestamp" in output_df.columns and "timestamp" not in datetimecols:
        datetimecols.append("timestamp")
    for col in datetimecols:
        output_df[col] = output_df[col].astype(dtype="datetime64[ns]", errors="ignore")
        output_df.loc[:, col] = pd.to_datetime(output_df[col], errors="ignore")

    datetimecols = output_df.select_dtypes(include=["datetime64"]).columns.tolist()
    formatters = {col: _format_date for col in datetimecols}
    return output_df.to_string(
        float_format="{:,.2f}".format,
        formatters=formatters,
        na_rep="None",
    )


def _format_date(date: pd.Timestamp):
    """
    Formats a pandas Timestamp object to a string, using the format "dd MMM YYYY HH:MM"

    Args:
        date (pd.Timestamp): The pandas Timestamp object to be formatted.

    Returns:
        str: The formatted string representation of the Timestamp object.
    """
    return date.strftime("%d %b %Y %H:%M")


def logging_decorator(logger: logging.Logger):
    """
    A decorator that logs the execution of a function using a specified logger.

    This decorator logs the start and completion of the function execution, including the
    input arguments and the result. If an exception occurs during the function execution,
    it logs the error along with the traceback.

    Args:
        logger (logging.Logger): The logger instance to be used for logging.

    Returns:
        function: The decorated function with added logging functionality.

    Example:
        import logging
        logger = logging.getLogger(__name__)
        @logging_decorator(logger)
        def example_function(x, y):
            return x + y
    """

    def decorator_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            logger.info(
                f"Starting {func.__name__}",
                extra={
                    "function": func.__name__,
                    "input_args": None if len(args) == 0 else args,
                    "input_kwargs": kwargs,
                },
            )
            try:
                result = func(*args, **kwargs)
                logger.info(
                    f"Completed {func.__name__}",
                    extra={
                        "function": func.__name__,
                        "input_args": None if len(args) == 0 else args,
                        "input_kwargs": kwargs,
                        "response": result,
                    },
                )
            except Exception as e:
                logger.error(
                    f"Error in {func.__name__}: {e.__class__.__name__}.",
                    extra={
                        "function": func.__name__,
                        "traceback": traceback.format_exc().splitlines(),
                    },
                )
                raise e
            return result

        return wrapper

    return decorator_wrapper



"""
PythonicAnalysisFactory class - generate and execute Pythonic code from text inputs.
"""

# standard library imports
import re
import logging
import warnings
from typing import Union

# third-party imports
import numpy as np
import pandas as pd
import pmdarima as pm
from scipy import stats
import statsmodels.api as sm

# local imports
from lyzr.base.llm import LiteLLM
from lyzr.base.errors import DependencyError
from lyzr.base.prompt import LyzrPromptFactory
from lyzr.base.base import UserMessage, SystemMessage
from lyzr.data_analyzr.models import FactoryBaseClass
from lyzr.data_analyzr.analysis_handler.utils import (
    extract_df_names,
    iterate_llm_calls,
    make_locals_string,
    extract_python_code,
    extract_column_names,
    handle_analysis_output,
    remove_print_and_plt_show,
)
from lyzr.data_analyzr.vector_store_utils import ChromaDBVectorStore


class PythonicAnalysisFactory(FactoryBaseClass):
    """
    PythonicAnalysisFactory is a class for generating and executing Pythonic analysis based on user input.

    Attributes:
        df_dict (dict): Dictionary containing dataframes for analysis.

    Methods:
        __init__(llm, logger, context, df_dict, vector_store, max_retries=None, time_limit=None, auto_train=None, **llm_kwargs):
            Initializes a PythonicAnalysisFactory instance.

        generate_output(user_input, **kwargs):
            Runs analysis and generates output based on the provided user input.

        get_prompt_messages(user_input):
            Generates a list of prompt messages based on the user's input.

        _get_locals_and_docs(system_message_sections, system_message_dict, user_input):
            Retrieves local variables and related documentation based on user input.

        extract_and_execute_code(llm_response):
            Extracts Python code from a given LLM response, processes it, and executes it within a controlled environment.

        code_cleaner(code) -> str:
            Handler for cleaning the extracted code.

        auto_train(user_input, code, **kwargs):
            Adds the user input and generated Python code to the vector store if the auto_train flag is set.
    """

    def __init__(
        self,
        llm: LiteLLM,
        logger: logging.Logger,
        context: str,
        df_dict: dict,
        vector_store: ChromaDBVectorStore,
        max_retries: int = None,
        time_limit: int = None,
        auto_train: bool = None,
        **llm_kwargs,
    ):
        """
        Initialize a PythonicAnalysisFactory instance.

        Args:
            llm (LiteLLM): The llm instance to be used.
            logger (logging.Logger): Logger instance for logging information.
            context (str): The context for the given query. If empty, pass "".
            df_dict (dict): Dictionary containing dataframes for analysis.
            vector_store (ChromaDBVectorStore): The vector store for managing related queries and database documentation.
            max_retries (int, optional): Maximum number of retries for analysis. Defaults to 10 if not provided.
            time_limit (int, optional): Time limit for analysis in seconds. Defaults to 45 if not provided.
            auto_train (bool, optional): Whether to automatically add to training data. Defaults to True.
            **llm_kwargs: Additional keyword arguments for the llm.

        Example:
            from lyzr.base.llm import LiteLLM
            from lyzr.data_analyzr.vector_store_utils import ChromaDBVectorStore
            from lyzr.data_analyzr.analysis_handler import PythonicAnalysisFactory
            import logging
            import pandas as pd

            llm = LiteLLM.from_defaults(model="gpt-4o")
            logger = logging.getLogger(__name__)
            context = "Context for the analysis."
            df_dict = {"df": pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})}
            vector_store = ChromaDBVectorStore(path="path/to/vector_store", logger=logger)
            analysis_factory = PythonicAnalysisFactory(
                llm=llm,
                logger=logger,
                context=context,
                df_dict=df_dict,
                vector_store=vector_store
            )
        """
        super().__init__(
            llm=llm,
            logger=logger,
            context=context,
            vector_store=vector_store,
            max_retries=10 if max_retries is None else max_retries,
            time_limit=45 if time_limit is None else time_limit,
            auto_train=auto_train,
            llm_kwargs=llm_kwargs,
        )
        self.df_dict = df_dict

    def generate_output(
        self, user_input: str, **kwargs
    ) -> Union[str, pd.DataFrame, dict[str, pd.DataFrame], None]:
        """
        Run analysis and generate output based on the provided user input.

        This is the primary method for generating and executing Pythonic analysis.
        - Processes the user input to generate Pythonic analysis output.
        - Handles retries and time limits for the code extraction and execution process.
        - Uses iterate_llm_calls to handle retries and time limits.

        Example:
            output = analysis_factory.generate_output("Your question here.")
            print(output)

        Args:
            user_input (str): The input string provided by the user for analysis.
            **kwargs: Additional keyword arguments to customize the analysis process.
                - max_retries (int): Maximum number of retries for LLM calls. Defaults to the value in self.params.max_retries.
                - time_limit (int): Time limit for the analysis execution. Defaults to the value in self.params.time_limit.
                - auto_train (bool): Whether to automatically add the analysis to training data. Defaults to the value in self.params.auto_train.

        Returns:
            Union[str, pd.DataFrame, dict[str, pd.DataFrame], None]: The output of the analysis.

        Raises:
            DependencyError: If the user input is None or an empty string.
        """
        user_input = user_input.strip()
        if user_input is None or user_input == "":
            raise DependencyError("A user input is required for analysis.")
        self.extract_and_execute_code = iterate_llm_calls(
            max_retries=kwargs.pop("max_retries", self.params.max_retries),
            llm=self.llm,
            llm_messages=self.get_prompt_messages(user_input),
            logger=self.logger,
            log_messages={
                "start": f"Starting Pythonic analysis for query: {user_input}",
                "end": f"Ending Pythonic analysis for query: {user_input}",
            },
            time_limit=kwargs.pop("time_limit", self.params.time_limit),
            llm_kwargs=dict(
                max_tokens=2000,
                top_p=1,
            ),
        )(self.extract_and_execute_code)
        self.output = handle_analysis_output(self.extract_and_execute_code())
        self.auto_train(user_input, self.code, **kwargs)
        return self.output

    def get_prompt_messages(self, user_input: str) -> list:
        """
        Generate a list of prompt messages based on the user's input.

        This method constructs a series of messages to be used with the LLM.
        - Incorporates context and examples relevant to the user's input.
        - List of messages includes system messages with context, local variables,
        documentation, and historical examples, followed by the user's input.

        Args:
            user_input (str): The input provided by the user.

        Returns:
            list[ChatMessage]: A list of chat messages generated based on the user's input.
        """
        system_message_sections = [
            "context",
            "external_context",
            "task",
            "closing",
        ]
        system_message_dict = {"context": self.context}
        # add locals and docs
        system_message_sections, system_message_dict = self._get_locals_and_docs(
            system_message_sections=system_message_sections,
            system_message_dict=system_message_dict,
            user_input=user_input,
        )
        # add question examples
        question_examples_list = self.vector_store.get_related_python_code(user_input)
        if len(question_examples_list) > 0:
            system_message_sections.append("history")
        messages = [
            LyzrPromptFactory(name="analysis_code", prompt_type="system").get_message(
                use_sections=system_message_sections,
                **system_message_dict,
            )
        ]
        for example in question_examples_list:
            if (
                (example is not None)
                and ("question" in example)
                and ("python_code" in example)
            ):
                messages.append(UserMessage(content=example["question"]))
                messages.append(SystemMessage(content=example["python_code"]))
        messages.append(UserMessage(content=user_input))
        return messages

    def _get_locals_and_docs(
        self, system_message_sections: list, system_message_dict: dict, user_input: str
    ) -> tuple[list, dict]:
        """
        Retrieve local variables and related documentation based on user input.

        Args:
            system_message_sections (list): A list to append sections of system messages.
            system_message_dict (dict): A dictionary to store system message format strings.
            user_input (str): The user input string to search for related documentation.

        Returns:
            tuple[list, dict]: Updated system message sections and system message dictionary.

        Procedure:
            - Retrieve related documentation from the vector store based on the user input.
            - Extract dataframe names from the documentation and ensure they exist in the local scope.
            - Update the local scope with dataframes that have been cleaned of NaN values.
            - Update the system message dictionary with the local variables and related documentation.
        """
        self.locals_ = {
            "pd": pd,
            "np": np,
            "pm": pm,
            "sm": sm,
            "stats": stats,
        }
        system_message_sections.append("locals")
        doc_list = self.vector_store.get_related_documentation(user_input)
        df_names = set()
        if len(doc_list) > 0:
            system_message_sections.append("doc_addition_text")
            doc_str = ""
            for doc_item in doc_list:
                doc_str += f"{doc_item}\n"
            for name in re.finditer(r"(\w+) dataframe", doc_str, re.IGNORECASE):
                df_names.update(name.groups())
            system_message_dict["doc"] = doc_str
        for name, df in self.df_dict.items():
            if name in df_names:
                self.locals_[name] = df
        system_message_dict["locals"] = make_locals_string(self.locals_)
        return system_message_sections, system_message_dict

    def extract_and_execute_code(self, llm_response: str):
        """
        Extracts Python code from a given LLM response, processes it, and executes it within a controlled environment.
        To be used as a callback function for iterate_llm_calls.

        Args:
            llm_response (str): The response from the language model containing Python code.

        Returns:
            Any: The result of the executed code, expected to be stored in the 'result' variable within the executed code's local scope.

        Raises:
            AssertionError: If any of the dataframes in `df_dict` are not instances of `pd.DataFrame`.

        Logs:
            - Info: The extracted Python code.

        Procedure:
            - Extracts Python code from the LLM response and removes any `print`, `plt.show()`, `plt.savefig()` statements.
            - Logs the extracted code.
            - Identifies dataframe names used in the code and ensures they exist in `df_dict`.
            - For each identified dataframe, drops rows with NaN values in the columns used in the code and updates the local scope.
            - Suppresses pandas chained assignment warnings and other warnings.
            - Executes the processed code within the local scope.
            - Stores the executed code in the `code` attribute.
            - Returns the value of the 'result' variable from the local scope.
        """
        code = self.code_cleaner(extract_python_code(llm_response))
        self.logger.info(f"Extracted Python code:\n{code}")
        df_names = extract_df_names(code, list(self.df_dict.keys()))
        for name in df_names:
            df = self.df_dict[name]
            assert isinstance(
                df, pd.DataFrame
            ), "df_dict must must only contain pandas DataFrames"
            columns = extract_column_names(code, self.df_dict[name])
            self.locals_[name] = df.dropna(subset=columns)
        pd.options.mode.chained_assignment = None
        warnings.filterwarnings("ignore")
        globals_ = self.locals_
        exec(code, globals_, self.locals_)
        self.code = code
        return self.locals_["result"]

    def code_cleaner(self, code: str) -> str:
        """Handler for cleaning the extracted code before execution."""
        return remove_print_and_plt_show(code)

    def auto_train(self, user_input: str, code: str, **kwargs):
        """
        Adds the user input and generated Python code to the vector store if the auto_train flag is set.

        Args:
            user_input (str): The input provided by the user to be used for training.
            code (str): The Python code generated by the analysis to be used for training.
            auto_train (bool): Optional flag to control whether the training should be performed. Defaults to the instance's 'auto_train' parameter.

        Procedure:
            - Checks if the 'auto_train' flag is set either in the keyword arguments or in the instance parameters.
            - Checks if the 'output' and 'code' attributes are not None and the 'code' attribute is not an empty string.
            - Adds a training plan to the vector store using the provided user input and the current Python code.
        """
        if (
            kwargs.pop("auto_train", self.params.auto_train)
            and self.output is not None
            and code is not None
            and code.strip() != ""
        ):
            self.vector_store.add_training_data(question=user_input, python_code=code)



"""
Utility functions for handling data, including reading files and directories.
"""

# standard library imports
import os
import io
import pickle
import logging
from typing import Union, Optional

# third-party imports
import pandas as pd
from pydantic import BaseModel

# local imports
from lyzr.data_analyzr.db_models import (
    SupportedDBs,
    DataFile,
    FilesConfig,
    RedshiftConfig,
    PostgresConfig,
    SQLiteConfig,
    DataConfig,
)
from lyzr.data_analyzr.db_connector import (
    DatabaseConnector,
    SQLiteConnector,
    TrainingPlan,
    TrainingPlanItem,
)
from lyzr.data_analyzr.models import AnalysisTypes
from lyzr.data_analyzr.models import VectorStoreConfig
from lyzr.data_analyzr.vector_store_utils import ChromaDBVectorStore
from lyzr.data_analyzr.utils import deterministic_uuid, translate_string_name


def get_db_details(
    analysis_type: AnalysisTypes,
    db_type: SupportedDBs,
    db_config: BaseModel,
    vector_store_config: VectorStoreConfig,
    logger: logging.Logger,
) -> tuple[DatabaseConnector, dict[str, pd.DataFrame], ChromaDBVectorStore]:
    """
    Retrieve database details including dataframes, connector, and vector store based on the provided configurations.

    Args:
        analysis_type (AnalysisTypes): The type of analysis to be performed.
        db_type (SupportedDBs): The type of database being used.
        db_config (BaseModel): Configuration for the database, must be an instance of
            FilesConfig, RedshiftConfig, PostgresConfig, or SQLiteConfig.
        vector_store_config (VectorStoreConfig): Configuration for the vector store.
        logger (logging.Logger): Logger instance for logging information.

    Returns:
        tuple: A tuple containing:
            - connector: The database connector instance.
            - df_dict (dict): A dictionary of dataframes read from the datasets.
            - vector_store (ChromaDBVectorStore): The vector store instance.

    Raises:
        AssertionError: If db_config is not an instance of the expected configuration class based on db_type.
    """
    df_dict = None
    connector = None
    vector_store = None
    # Read given datasets
    if db_type is SupportedDBs.files:
        assert isinstance(
            db_config, FilesConfig
        ), f"Expected FilesConfig, got {type(db_config)}"
        df_dict = get_dict_of_files(db_config.datasets)
        logger.info(
            "Following datasets read successfully:\n"
            + "\n".join([f"{df} with shape {df_dict[df].shape}" for df in df_dict])
            + "\n"
        )
    else:
        accepted_db_types = tuple(
            [
                elem
                for elem in DataConfig._config_types.values()
                if elem is not FilesConfig
            ]
        )
        assert isinstance(
            db_config, accepted_db_types
        ), f"Expected one of {accepted_db_types}, got {type(db_config)}"
        connector = DatabaseConnector.get_connector(db_type)(**db_config.model_dump())
    df_dict, connector = ensure_correct_data_format(
        analysis_type=analysis_type,
        db_config=db_config,
        df_dict=df_dict,
        connector=connector,
    )
    # Create vector store
    if vector_store_config.path is None:
        uuid = deterministic_uuid(
            [
                analysis_type.value,
                db_type.value,
                " ".join([f"{k}: {v}" for k, v in db_config.model_dump().items()]),
            ]
        )
        vector_store_config.path = f"./vector_store/{uuid}"
    vector_store = ChromaDBVectorStore(
        path=vector_store_config.path,
        remake_store=vector_store_config.remake_store,
        training_plan_func=make_training_plan,
        logger=logger,
        analysis_type=analysis_type,
        db_type=db_type,
        df_dict=df_dict,
        connector=connector,
    )
    return connector, df_dict, vector_store


def ensure_correct_data_format(
    analysis_type: AnalysisTypes,
    db_config: Union[FilesConfig, RedshiftConfig, PostgresConfig, SQLiteConfig],
    df_dict: Optional[dict[str, pd.DataFrame]] = None,
    connector: Optional[DatabaseConnector] = None,
) -> tuple[dict[str, pd.DataFrame], DatabaseConnector]:
    """
    Ensure the correct format of data based on the analysis type and provided configurations.

    This function adjusts the format of the data (either as pandas DataFrames or a SQL connector)
    depending on the specified analysis type.
    - For pythonic analysis, it ensures data is in the form of pandas DataFrames.
    - For SQL analysis, it ensures data is accessible via a SQL connector.

    Args:
        analysis_type (AnalysisTypes): The type of analysis to be performed (pythonic, SQL, or neither).
        db_config (Union[FilesConfig, RedshiftConfig, PostgresConfig, SQLiteConfig]): Configuration for the database.
        df_dict (Optional[dict[str, pd.DataFrame]], optional): A dictionary of dataframes. Defaults to None.
        connector (Optional[DatabaseConnector], optional): A database connector instance. Defaults to None.

    Returns:
        tuple: A tuple containing:
            - df_dict (Optional[dict[str, pd.DataFrame]]): The dictionary of dataframes, if applicable.
            - connector (Optional[DatabaseConnector]): The database connector instance, if applicable.
    """
    if analysis_type is AnalysisTypes.ml and df_dict is None:
        df_dict = connector.fetch_dataframes_dict()
        connector = None
    if df_dict is not None:
        df_dict = {translate_string_name(k): v for k, v in df_dict.items()}
    if analysis_type is AnalysisTypes.sql and connector is None:
        connector = SQLiteConnector()
        connector.create_database(
            db_path=(
                db_config.db_path
                if db_config.db_path is not None
                else f"./sqlite/{deterministic_uuid()}.db"
            ),
            df_dict=df_dict,
        )
        df_dict = None
    return df_dict, connector


def make_training_plan(
    analysis_type: AnalysisTypes,
    db_type: SupportedDBs,
    df_dict: dict,
    connector: DatabaseConnector,
) -> TrainingPlan:
    """
    Create a training plan based on the analysis type, database type, and provided data.

    This function generates a training plan for pythonic or SQL analysis based on the specified
    analysis type and database type. It processes dataframes or uses a database connector to
    create the appropriate training plan.

    Args:
        analysis_type (AnalysisTypes): The type of analysis to be performed (Pythonic, SQL, or skip).
        db_type (SupportedDBs): The type of database being used (files, Redshift, Postgres, SQLite).
        df_dict (dict): A dictionary of dataframes to be used for the training plan.
        connector (DatabaseConnector): The database connector instance for SQL databases.

    Returns:
        TrainingPlan: The generated training plan based on the provided data and configurations.

    Raises:
        AssertionError: If the dataframes in df_dict are not instances of pandas DataFrame.
        AssertionError: If the training plan could not be created from the given data.
    """
    training_plan = None
    if analysis_type is AnalysisTypes.ml or (
        (analysis_type is AnalysisTypes.skip) and (db_type is SupportedDBs.files)
    ):
        training_plan = TrainingPlan([])
        for name, df in df_dict.items():
            assert isinstance(df, pd.DataFrame), f"Expected DataFrame, got {type(df)}"
            doc = f"The following columns are in the {name} dataframe:\n\n"
            buffer = io.StringIO()
            df.info(buf=buffer)
            doc += buffer.getvalue()
            training_plan._plan.append(
                TrainingPlanItem(
                    item_type=TrainingPlanItem.ITEM_TYPE_IS,
                    item_group=name,
                    item_name=name,
                    item_value=doc,
                )
            )
            doc = f"Following are the first five rows of the {name} dataframe:\n\n"
            doc += df.head().to_markdown()
            training_plan._plan.append(
                TrainingPlanItem(
                    item_type=TrainingPlanItem.ITEM_TYPE_IS,
                    item_group=name,
                    item_name=name,
                    item_value=doc,
                )
            )
    elif analysis_type is AnalysisTypes.sql or (
        (analysis_type is AnalysisTypes.skip) and (db_type is not SupportedDBs.files)
    ):
        training_plan = connector.get_default_training_plan()
    assert isinstance(
        training_plan, TrainingPlan
    ), "Unable to create training plan from given data."
    return training_plan


def get_dict_of_files(datasets: list[DataFile]) -> dict[str, pd.DataFrame]:
    """
    Retrieve a dictionary of dataframes from a list of data files.

    Args:
        datasets (list[DataFile]): A list of DataFile objects, each containing the name, value,
            and keyword arguments for reading the file.

    Returns:
        dict[str, pd.DataFrame]: A dictionary where the keys are the formatted names of the data
            files and the values are the corresponding dataframes.

    Precedure:
    - Process a list of data files
    - Read each file, folder, or dataframe
    - Store the resulting dataframes in a dictionary
    - The keys of the dictionary are formatted to
        - replaces spaces and punctuation with underscores (_)
        - converts all characters to lowercase
    """
    datasets_dict = {}
    for data in datasets:
        datasets_dict.update(read_file_or_folder(data.name, data.value, data.kwargs))
    df_names = list(datasets_dict.keys())
    for name in df_names:
        new_df_name = translate_string_name(name)
        datasets_dict[new_df_name] = datasets_dict.pop(name)
        rename_cols = {
            col: translate_string_name(col)
            for col in datasets_dict[new_df_name].columns
        }
        datasets_dict[new_df_name].rename(columns=rename_cols, inplace=True)
    return datasets_dict


def read_file_or_folder(
    name: str, filepath: Union[str, pd.DataFrame], kwargs
) -> dict[str, pd.DataFrame]:
    """
    Reads a file or all files in a directory and returns their contents as a dictionary of DataFrames.

    This function handles both individual files and directories:
    - If filepath is a DataFrame, it returns a dictionary with the given name as the key.
    - If filepath is a file, it reads the file and returns a dictionary with the given name as the key.
    - If filepath is a directory, it reads all files in the directory (excluding nested directories) and
      returns a dictionary with keys derived from the filenames.

    Args:
        name (str): The base name to use for the dictionary keys.
        filepath (Union[str, pd.DataFrame]): The path to the file or directory, or a DataFrame.
        kwargs: Additional keyword arguments to pass to the `read_file` function.

    Returns:
        dict[str, pd.DataFrame]: A dictionary where keys are derived from the provided name and filenames, and values are DataFrames.

    Raises:
        ValueError: If filepath is neither a valid file nor a directory.
    """
    if isinstance(filepath, pd.DataFrame):
        return {name: filepath}
    if os.path.isfile(filepath):
        return {name: read_file(filepath, **kwargs)}
    elif os.path.isdir(filepath):
        return {
            f"{name}_{os.path.splitext(file)[0]}": read_file(
                os.path.join(filepath, file), **kwargs
            )
            for file in os.listdir(filepath)
            if os.path.isfile(
                os.path.join(filepath, file)
            )  # do not read from nested directories
        }

    else:
        raise ValueError(
            f"Filepath '{filepath}' is not a valid file or directory. Please provide a valid filepath."
        )


def read_file(filepath: str, **kwargs) -> pd.DataFrame:
    """
    Read a file and return its contents as a pandas DataFrame.

    This function reads a file from the specified filepath and returns its contents as a pandas DataFrame.
    Supported file formats include CSV, TSV, JSON, Excel, TXT, and pickle files.

    Args:
        filepath (str): The path to the file to be read.
        encoding (Optional[str], optional): The encoding of the file. Defaults to "utf-8".
        **kwargs: Additional keyword arguments to pass to the pandas read function.

    Returns:
        pd.DataFrame: The contents of the file as a pandas DataFrame.

    Raises:
        FileNotFoundError: If the file is not found at the specified filepath.
        UnicodeDecodeError: If the file could not be decoded with the specified encoding.
        ValueError:
            - If the file does not exist
            - If the file extension is not supported
            - If the file could not be read as a DataFrame.
    """
    if not os.path.exists(filepath):
        raise ValueError(
            f"File '{filepath}' not found. Please provide a valid filepath."
        )
    file_extension = filepath.split(".")[-1]
    try:
        if file_extension == "csv" or file_extension == "txt":
            df = pd.read_csv(filepath, **kwargs)
        elif file_extension == "tsv":
            df = pd.read_csv(filepath, sep="\t", **kwargs)
        elif file_extension == "json":
            df = pd.read_json(filepath, **kwargs)
        elif file_extension in ["xlsx", "xls"]:
            df = pd.read_excel(filepath, **kwargs)
        elif file_extension == "pkl":
            with open(filepath, "rb") as f:
                df = pickle.load(f, **kwargs)
        else:
            raise ValueError(
                f"File extension '{file_extension}' not supported. Please provide a csv or pkl file."
            )
    except FileNotFoundError:
        raise FileNotFoundError(
            f"File '{filepath}' not found. Please provide a valid filepath."
        )
    except UnicodeDecodeError:
        raise UnicodeDecodeError(
            f"File '{filepath}' could not be decoded. Please provide a file with utf-8 encoding."
            "If the file is not encoded in utf-8, please provide the encoding as a parameter: file_kwargs={'encoding': 'utf-8'}"
        )
    if not isinstance(df, pd.DataFrame):
        raise ValueError(
            f"File '{filepath}' could not be read as a DataFrame. Please provide a valid file."
        )
    return df



"""
PlotFactory class - generate visualizations based on user input and data sources.
"""

# standard library imports
import os
import re
import time
import logging
import warnings
import traceback

# third-party imports
import numpy as np
import pandas as pd
import seaborn as sns
import pmdarima as pm
from scipy import stats
import statsmodels.api as sm
import matplotlib.pyplot as plt

# local imports
from lyzr.base.llm import LiteLLM
from lyzr.base.prompt import LyzrPromptFactory
from lyzr.base.base import UserMessage, SystemMessage
from lyzr.data_analyzr.models import FactoryBaseClass, AnalysisTypes
from lyzr.data_analyzr.utils import deterministic_uuid
from lyzr.data_analyzr.analysis_handler.utils import (
    handle_plotpath,
    extract_df_names,
    iterate_llm_calls,
    make_locals_string,
    extract_python_code,
    extract_column_names,
    remove_print_and_plt_show,
)
from lyzr.data_analyzr.db_connector import DatabaseConnector
from lyzr.data_analyzr.vector_store_utils import ChromaDBVectorStore

default_plot_path = "generated_plots/plot.png"


class PlotFactory(FactoryBaseClass):
    """
    PlotFactory is a class responsible for generating visual representations based on user input.

    This class leverages a language model (LLM) to interpret user queries and generate corresponding plots.
    It supports various data sources, including database connectors and dictionaries of pandas DataFrames.
    The class also manages retries, time limits, and automatic training of the vector store with new plotting examples.

    Attributes:
        plotting_library (str): The plotting library used, default is "matplotlib".
        connector (DatabaseConnector): The database connector for fetching data, if provided.
        df_dict (dict): Dictionary of pandas DataFrames, if provided.
        analysis_output (pd.DataFrame or dict): The output of analysis, if provided.
        plot_path (str): Path where the plot will be saved.
        locals_ (dict): Dictionary of local variables for use in plotting functions.
        code (str): The executed plotting code.
        fig (matplotlib.figure.Figure): The generated plot figure.

    Methods:
        generate_output(user_input: str, plot_path: str = None, **kwargs) -> str:
            Generates a visual representation based on the provided user input.

        get_prompt_messages(user_input: str) -> list:
            Generate a list of prompt messages based on the user's input.

        _get_message_sections_and_dict(user_input: str) -> tuple[list, dict]:
            Extracts and organizes message sections and a corresponding dictionary based on user input.

        _get_message_docs(user_input: str) -> tuple[str, set]:
            Retrieve related documentation and dataframe names using the user input.

        _get_locals() -> dict:
            Retrieve a dictionary of local variables for use in plotting functions.

        _add_sql_examples(user_input: str, system_message_sections: list, system_message_dict: dict):
            Add SQL examples to the system message sections and dictionary based on user input.

        _add_python_examples(user_input: str, system_message_sections: list, system_message_dict: dict):
            Add Python examples to the system message sections and dictionary based on user input.

        extract_and_execute_code(llm_response: str):
            Executes the plotting code extracted from the provided LLM response.

        code_cleaner(code: str) -> str:
            Handler for cleaning the code by removing print statements and plt.show() calls.

        save_plot_image() -> str:
            Saves the current plot to a file specified by `self.plot_path`.

        auto_train(user_input: str, code: str, **kwargs):
            Automatically trains the vector store with a given user input and corresponding plot code.
    """

    def __init__(
        self,
        llm: LiteLLM,
        logger: logging.Logger,
        context: str,
        analysis_type: AnalysisTypes,
        data_kwargs: dict,
        vector_store: ChromaDBVectorStore,
        max_retries: int = None,
        time_limit: int = None,
        auto_train: bool = None,
        **llm_kwargs,  # model_kwargs: dict
    ):
        """
        Initializes the PlotFactory instance.

        Args:
            llm (LiteLLM): The llm instance to be used.
            logger (logging.Logger): Logger instance for logging purposes.
            context (str): The context for the given query. If empty, pass "".
            analysis_type (AnalysisTypes): The type of analysis performed.
            data_kwargs (dict): Dictionary containing data-related parameters.
                - connector (DatabaseConnector, optional): Database connector for fetching data.
                - df_dict (dict, optional): Dictionary of pandas DataFrames.
                - analysis_output (optional): Output of analysis.
            vector_store (ChromaDBVectorStore): The vector store for managing related queries and database documentation.
            max_retries (int, optional): Maximum number of retries for plotting. Defaults to 10 if not provided.
            time_limit (int, optional): Time limit for plotting in seconds. Defaults to 60 if not provided.
            auto_train (bool, optional): Whether to automatically add to training data. Defaults to True.
            **llm_kwargs: Additional keyword arguments for the language model.

        Raises:
            ValueError: If neither `connector` nor `df_dict` is provided.
            ValueError: If `connector` is not a DatabaseConnector instance or `df_dict` is not a dictionary of pandas DataFrames.

        Example:
            from lyzr.base.llm import LiteLLM
            from lyzr.data_analyzr.vector_store_utils import ChromaDBVectorStore
            from lyzr.data_analyzr.analysis_handler import PlotFactory
            llm = LiteLLM.from_defaults(model="gpt-4o")
            logger = logging.getLogger('plot_logger')
            context = "Provide context for the plot"
            data_kwargs = {"df_dict": {"example_df": pd.DataFrame(...)}}
            vector_store = ChromaDBVectorStore(path="path/to/vector_store")
            plotter = PlotFactory(
                llm=llm,
                logger=logger,
                context=context,
                analysis_type=AnalysisTypes.ml,
                data_kwargs=data_kwargs,
                vector_store=vector_store,
                max_retries=10, # default
                time_limit=60, # default
                auto_train=True
            )
        """
        super().__init__(
            llm=llm,
            logger=logger,
            context=context,
            vector_store=vector_store,
            max_retries=10 if max_retries is None else max_retries,
            time_limit=60 if time_limit is None else time_limit,
            auto_train=auto_train,
            llm_kwargs=llm_kwargs,
        )
        self.analysis_type = analysis_type
        self.plotting_library = "matplotlib"
        self.connector = data_kwargs.get("connector", None)
        self.df_dict = data_kwargs.get("df_dict", None)
        self.analysis_output = data_kwargs.get("analysis_output", None)
        if self.connector is None and self.df_dict is None:
            raise ValueError(
                "Either connector or df_dict must be provided to make a plot."
            )
        if not isinstance(self.connector, DatabaseConnector) and not isinstance(
            self.df_dict, dict
        ):
            raise ValueError(
                "Either connector or df_dict must be a DatabaseConnector or a dictionary of pandas DataFrames."
            )

    def generate_output(self, user_input: str, plot_path: str = None, **kwargs) -> str:
        """
        Generates a visual representation based on the provided user input.

        This is the primary method for generating plots.
        - Processes the user input to generate visualisation.
        - Handles retries and time limits for the code extraction and execution process.
        - Uses iterate_llm_calls to handle retries and time limits.

        Args:
            user_input (str): The input string from the user that describes the desired plot.
            plot_path (str): Path where the plot will be saved.
            **kwargs: Additional keyword arguments to customize the behavior of the method.
                - max_retries (int): Maximum number of retries for LLM calls. Defaults to self.params.max_retries.
                - time_limit (int): Time limit for LLM calls. Defaults to self.params.time_limit.
                - auto_train (bool): Whether to automatically add the generated code to the training data. Defaults to self.params.auto_train.

        Returns:
            str: The file path of the saved plot image, or an empty string if no plot was generated.

        Example:
            saved_plot_path = plotter.get_visualisation("Plot a bar chart of sales by region")
            from PIL import Image
            Image.open(plot_path).show()
        """
        self.plot_path = handle_plotpath(
            plot_path=plot_path,
            output_format="png",
            uuid=deterministic_uuid([user_input, str(time.time())]),
            logger=self.logger,
        )
        messages = self.get_prompt_messages(user_input)
        self.extract_and_execute_code = iterate_llm_calls(
            max_retries=kwargs.pop("max_retries", self.params.max_retries),
            llm=self.llm,
            llm_messages=messages,
            logger=self.logger,
            log_messages={
                "start": f"Generating plot for query: {user_input}",
                "end": f"Finished generating plot for query: {user_input}",
            },
            time_limit=kwargs.pop("time_limit", self.params.time_limit),
            llm_kwargs=dict(
                max_tokens=2000,
                top_p=1,
            ),
        )(self.extract_and_execute_code)
        self.fig = self.extract_and_execute_code()
        if self.fig is None:
            plt.close("all")
            return ""
        self.auto_train(user_input, self.code, **kwargs)
        return self.save_plot_image()

    def get_prompt_messages(self, user_input: str) -> list:
        """
        Generate a list of prompt messages based on the user's input.

        This method processes the user's input to create a series of messages that
        can be used to generate a plot. It retrieves message sections and a dictionary
        of system messages, fetches similar plotting code examples from the vector
        store, and appends them to the message list. Finally, it constructs and returns
        a list of messages, including the user's input.

        Args:
            user_input (str): The input provided by the user.

        Returns:
            list: A list of messages including system prompt, examples, and user input.
        """
        system_message_sections, system_message_dict = (
            self._get_message_sections_and_dict(user_input=user_input)
        )
        # add question examples
        question_examples_list = self.vector_store.get_related_plotting_code(user_input)
        if len(question_examples_list) > 0:
            system_message_sections.append("history")
        messages = [
            LyzrPromptFactory(name="plotting_code", prompt_type="system").get_message(
                use_sections=system_message_sections,
                **system_message_dict,
            )
        ]
        for example in question_examples_list:
            if (
                (example is not None)
                and ("question" in example)
                and ("plot_code" in example)
            ):
                messages.append(UserMessage(content=example["question"]))
                messages.append(SystemMessage(content=example["plot_code"]))
        messages.append(UserMessage(content=user_input))
        return messages

    def _get_message_sections_and_dict(self, user_input: str) -> tuple[list, dict]:
        """
        Extracts and organizes message sections and a corresponding dictionary based on user input.

        Args:
            user_input (str): The input provided by the user.

        Returns:
            tuple[list, dict]: A tuple containing a list of message sections and a dictionary with context and data.

        Procedure:
            - Get related documentation and DataFrame names from _get_message_docs.
            - Get local variables from _get_locals.
            - If analysis_type is AnalysisTypes.sql, add SQL plot sections and examples.
            - If analysis_type is not AnalysisTypes.sql, add Python plot sections and examples.
            - Return the message sections and dictionary.
        """
        system_message_sections = ["context", "external_context"]
        system_message_dict = {"context": self.context}
        doc_str, df_names = self._get_message_docs(user_input)
        self.locals_ = self._get_locals()
        if self.analysis_type is AnalysisTypes.sql:
            system_message_sections.append("sql_plot")
            system_message_sections.append("locals")
            self.locals_["conn"] = self.connector
            if doc_str is not None:
                system_message_sections.append("doc_addition_text")
                system_message_dict["doc"] = doc_str
                system_message_dict["db_type"] = "sql database"
            system_message_sections, system_message_dict = self._add_sql_examples(
                user_input=user_input,
                system_message_sections=system_message_sections,
                system_message_dict=system_message_dict,
            )
        else:
            system_message_sections.append("python_plot")
            system_message_sections.append("locals")
            self.locals_.update(
                {name: df for name, df in self.df_dict.items() if name in df_names}
            )
            if doc_str is not None:
                system_message_sections.append("doc_addition_text")
                system_message_dict["doc"] = doc_str
                system_message_dict["db_type"] = "dataframe(s)"
            system_message_sections, system_message_dict = self._add_python_examples(
                user_input=user_input,
                system_message_sections=system_message_sections,
                system_message_dict=system_message_dict,
            )
        system_message_dict["locals"] = make_locals_string(self.locals_)
        return system_message_sections, system_message_dict

    def _get_message_docs(self, user_input: str) -> tuple[str, set]:
        """
        Retrieve related documentation and dataframe names using the user input.

        Args:
            user_input (str): The input string provided by the user to search for related documentation.

        Returns:
            tuple[str, set]: A tuple containing:
                - A concatenated string of related documentation if any are found, otherwise None.
                - A set of dataframe names extracted from the documentation if any are found, otherwise None.

        Procedure:
            - Get related documentation from the vector store.
            - Extract dataframe names from the documentation.
            - Return the concatenated documentation string and the set of dataframe names.
        """
        doc_list = self.vector_store.get_related_documentation(user_input)
        df_names = set()
        if len(doc_list) > 0:
            doc_str = ""
            for doc_item in doc_list:
                doc_str += f"{doc_item}\n"
            for name in re.finditer(r"(\w+) dataframe", doc_str, re.IGNORECASE):
                df_names.update(name.groups())
            return doc_str, df_names
        return None, None

    def _get_locals(self):
        """
        Retrieve a dictionary of local variables for use in plotting functions.

        Returns:
            dict: A dictionary with keys as module names and values as the corresponding
            imported modules or DataFrames from `analysis_output`.

        Procedure:
            - Construct a dictionary containing commonly used libraries and modules.
            - Include the `analysis_output` attribute if it is a pandas DataFrame.
            - Update the dictionary with items from `analysis_output` if it is a dictionary containing DataFrames.
            - Return the dictionary.
        """
        locals_ = {
            "pd": pd,
            "np": np,
            "pm": pm,
            "sm": sm,
            "stats": stats,
            "plt": plt,
            "sns": sns,
        }
        if isinstance(self.analysis_output, pd.DataFrame):
            locals_["analysis_output"] = self.analysis_output
        elif isinstance(self.analysis_output, dict):
            locals_.update(
                {
                    name: df
                    for name, df in self.analysis_output.items()
                    if isinstance(df, pd.DataFrame)
                }
            )
        return locals_

    def _add_sql_examples(
        self, user_input: str, system_message_sections: list, system_message_dict: dict
    ):
        """
        Add SQL examples to the system message sections and dictionary based on user input.

        Args:
            user_input (str): The input provided by the user.
            system_message_sections (list): A list of sections in the system message.
            system_message_dict (dict): A dictionary containing the system message content.

        Returns:
            tuple: Updated `system_message_sections` and `system_message_dict` with SQL examples included if any were found.

        Procedure:
            - Retrieve SQL examples similar to the user's input from the vector store.
            - If any examples are found, append them to the system message sections and format them into a string.
            - Add the formatted string to the system message dictionary under the key "sql_examples".
            - Return the updated system message sections and dictionary.
        """
        sql_examples = (
            self.vector_store.get_related_sql_queries(user_input)
            if isinstance(self.connector, DatabaseConnector)
            else []
        )
        if len(sql_examples) > 0:
            system_message_sections.append("sql_examples_text")
            sql_examples_str = ""
            for example in sql_examples:
                if example is not None:
                    if "question" in example and "sql" in example:
                        sql_examples_str += f"Question: {example['question']}\nSQL:\n{example['sql']}\n\n"
            system_message_dict["sql_examples"] = sql_examples_str
        return system_message_sections, system_message_dict

    def _add_python_examples(
        self, user_input: str, system_message_sections: list, system_message_dict: dict
    ):
        """
        Add SQL examples to the system message sections and dictionary based on user input.

        Args:
            user_input (str): The input provided by the user.
            system_message_sections (list): A list of sections in the system message.
            system_message_dict (dict): A dictionary containing the system message content.

        Returns:
            tuple: Updated `system_message_sections` and `system_message_dict` with SQL examples included if any were found.

        Procedure:
            - Retrieve SQL examples similar to the user's input from the vector store.
            - If any examples are found, append them to the system message sections and format them into a string.
            - Add the formatted string to the system message dictionary under the key "python_examples".
            - Return the updated system message sections and dictionary.
        """
        python_examples = self.vector_store.get_related_python_code(user_input)
        if len(python_examples) > 0:
            system_message_sections.append("python_examples_text")
            python_examples_str = ""
            for example in python_examples:
                if example is not None:
                    if "question" in example and "python_code" in example:
                        python_examples_str += f"Question: {example['question']}\nAnalysis Code:\n{example['python_code']}\n\n"
            system_message_dict["python_examples"] = python_examples_str
        return system_message_sections, system_message_dict

    def extract_and_execute_code(self, llm_response: str):
        """
        Executes the plotting code extracted from the provided LLM response.
        To be used as a callback function for iterate_llm_calls.

        Args:
            llm_response (str): The response from the language model containing the plotting code.

        Returns:
            matplotlib.figure.Figure: The generated plot figure.

        Raises:
            AssertionError: When `self.connector` is not an instance of DatabaseConnector and:
                - `df_dict` is not a dictionary
                - if `df_dict` contains non-pandas DataFrame objects.

        Logs:
            - Info: Extracted Python code.

        Procedure:
            - Extracts Python code from the LLM response and removes any `print`, `plt.show()`, `plt.savefig()` statements.
            - Logs the extracted code.
            - If a DatabaseConnector is not present:
                a. Identifies dataframe names used in the code and ensures they exist in `df_dict`.
                b. For each identified dataframe, drops rows with NaN values in the columns used in the code
                and updates the local scope.
            - Suppresses pandas chained assignment warnings and other warnings.
            - Executes the processed code within the local scope.
            - Stores the executed code in the `code` attribute.
            - Returns the value of the 'result' variable from the local scope.
        """
        code = self.code_cleaner(extract_python_code(llm_response))
        self.logger.info(f"Extracted Python code:\n{code}")
        if self.analysis_type is AnalysisTypes.ml:
            assert isinstance(self.df_dict, dict), "df_dict must be a dictionary."
            df_names = extract_df_names(code, list(self.df_dict.keys()))
            for name in df_names:
                df = self.df_dict[name]
                assert isinstance(
                    df, pd.DataFrame
                ), "df_dict must must only contain pandas DataFrames"
                columns = extract_column_names(code, self.df_dict[name])
                self.locals_[name] = df.dropna(subset=columns)
        pd.options.mode.chained_assignment = None
        warnings.filterwarnings("ignore")
        globals_ = self.locals_
        exec(code, globals_, self.locals_)
        self.code = code
        return self.locals_["fig"]

    def code_cleaner(self, code: str) -> str:
        """Handler for cleaning the extracted code before execution."""
        return remove_print_and_plt_show(code)

    def save_plot_image(self) -> str:
        """
        Saves the current plot to a file specified by `self.plot_path`.

        Returns:
            str: The path where the plot was saved or intended to be saved.

        Logs:
            - Info: When the plot is successfully saved.
            - Error: When there is an issue saving the plot, including the traceback for debugging.
        """
        plt.tight_layout()
        if not PlotFactory._savefig(self.plot_path):
            self.logger.error(
                f"Error saving plot at: {self.plot_path}. Plot not saved. Displaying plot instead. Access the plot using `.fig` attribute.",
                extra={
                    "function": "save_plot_image",
                    "traceback": traceback.format_exc().splitlines(),
                },
            )
            plt.show()
        else:
            self.logger.info(
                f"\nPlot saved at: {self.plot_path}\n",
                extra={"function": "save_plot_image"},
            )
            plt.close("all")
        return self.plot_path

    @staticmethod
    def _savefig(path: str):
        try:
            dir_path = os.path.dirname(path)
            if dir_path.strip() != "":
                os.makedirs(dir_path, exist_ok=True)
            plt.savefig(path)
            return True
        except Exception:
            if path == default_plot_path:
                return False
            PlotFactory._savefig(default_plot_path)
        return False

    def auto_train(self, user_input: str, code: str, **kwargs):
        """
        Automatically trains the vector store with a given user input and corresponding plot code.

        Args:
            user_input (str): The input provided by the user, typically a question or query.
            code (str): The plot code or script that corresponds to the user input.
            auto_train (bool): Whether to automatically add the analysis to the training data,
                defaults to self.params.auto_train.

        Procedure:
            - Checks if the `auto_train` keyword is set to True.
            - If the plot code is not empty, adds the user input and the plot code to the vector store.
        """
        if (
            kwargs.pop("auto_train", self.params.auto_train)
            and code is not None
            and code.strip() != ""
        ):
            self.vector_store.add_training_data(question=user_input, plot_code=code)



"""
Classes for connecting to various databases and executing SQL queries.
"""

# standard library imports
import os
import sqlite3
import requests
from typing import Union
from pathlib import Path
from dataclasses import dataclass
from urllib.parse import urlparse

# third-party imports
import pandas as pd

# local imports
from lyzr.base.errors import (
    ImproperlyConfigured,
    ValidationError,
    DependencyError,
)
from lyzr.data_analyzr.db_models import SupportedDBs

required_modules = {
    "redshift_connector": "redshift_connector==2.0.918",
    "mysql.connector": "mysql-connector-python==8.2.0",
    "snowflake.connector": "snowflake-connector-python==3.6.0",
    "psycopg2": "psycopg2-binary==2.9.9",
}


@dataclass
class TrainingPlanItem:
    """
    A class representing an item in a training plan.

    Attributes:
        item_type (str): The type of the training item (e.g., SQL, DDL, Information Schema).
        item_group (str): The group to which the training item belongs.
        item_name (str): The name of the training item.
        item_value (str): The value associated with the training item.

    Methods:
        __str__(): Returns a string representation of the training item based on its type.
    """

    item_type: str
    item_group: str
    item_name: str
    item_value: str

    def __str__(self):
        if self.item_type == self.ITEM_TYPE_SQL:
            return f"Train on SQL: {self.item_group} {self.item_name}"
        elif self.item_type == self.ITEM_TYPE_DDL:
            return f"Train on DDL: {self.item_group} {self.item_name}"
        elif self.item_type == self.ITEM_TYPE_IS:
            return f"Train on Information Schema: {self.item_group} {self.item_name}"

    ITEM_TYPE_SQL = "sql"
    ITEM_TYPE_DDL = "ddl"
    ITEM_TYPE_IS = "is"
    ITEM_TYPE_PY = "py"
    ITEM_TYPE_PLOT = "plot"


class TrainingPlan:
    """
    A class representing a training plan consisting of multiple training items.

    Attributes:
        _plan (list[TrainingPlanItem]): A list of training plan items.

    Methods:
        __init__(plan: list[TrainingPlanItem]): Initializes the training plan with a list of items.
        __str__(): Returns a string representation of the entire training plan.
        __repr__(): Returns a string representation of the entire training plan.
        get_summary() -> list[str]: Returns a summary of the training plan as a list of strings.
        remove_item(item: str): Removes a training item from the plan based on its string representation.

    Example:
        plan = TrainingPlan([
            TrainingPlanItem(
                item_type="sql",
                item_group="public",
                item_name="table_name",
                item_value="SELECT * FROM table_name;"
            )
        ])
    """

    _plan: list[TrainingPlanItem]

    def __init__(self, plan: list[TrainingPlanItem]):
        self._plan = plan

    def __str__(self):
        return "\n".join(self.get_summary())

    def __repr__(self):
        return self.__str__()

    def get_summary(self) -> list[str]:
        return [f"{item}" for item in self._plan]

    def remove_item(self, item: str):
        for plan_item in self._plan:
            if str(plan_item) == item:
                self._plan.remove(plan_item)
                break


def import_modules(modules: dict[str, str]):
    """
    Dynamically imports a list of modules with specified versions and returns the imported modules.

    Args:
        modules (dict[str, str]): A dictionary where keys are module names (str) and values are the required versions (str).

    Returns:
        list: A list of imported module objects.

    Raises:
        DependencyError: If a module cannot be imported, an exception is raised with the module name and version.

    Example:
        modules = {
            'numpy': '1.21.0',
            'pandas': '1.3.0'
        }
        imported_modules = import_modules(modules)
    """
    imported_modules = []
    for mod, ver in modules.items():
        try:
            globals()[mod] = __import__(mod)
        except ImportError:
            raise DependencyError({mod: ver})
        imported_modules.append(globals()[mod])
    return imported_modules


def register(name: str):
    """
    A decorator function to register a connector class with the DatabaseConnector class.

    Args:
        name (str): The name or type of the database connector class, e.g. "postgres", "mysql", "redshift".
    """

    def wrapper(cls):
        DatabaseConnector.register_connector(name, cls)
        return cls

    return wrapper


class DatabaseConnector:
    """Parent class for all database connectors."""

    def __init__(self):
        """Initialize a connection to a database using the input credentials."""

    def fetch_dataframes_dict(self, **kwargs):
        """
        Fetches data from specified tables in a database and returns them as a dictionary of pandas DataFrames.

        Args:
            schema (Union[str, list], optional):
                The schema(s) to fetch tables from. If not provided, defaults to the instance's schema attribute.
            tables (list[str], optional):
                The list of table names to fetch. If not provided, defaults to the instance's tables attribute.

        Returns:
            dict[pd.DataFrame]:
                A dictionary where the keys are table names and the values are pandas DataFrames containing the table data.

        Raises:
            RuntimeError: If there is no connection to the database.
            ValidationError: If an error occurs while connecting to the database.
            RuntimeError: If an error occurs while fetching data from the table.
        """
        raise NotImplementedError("You need to connect to a database first.")

    def run_sql(self, **kwargs):
        """
        Executes a given SQL query on the connected database and returns the results as a pandas DataFrame.

        Args:
            sql (str): The SQL query to be executed.

        Returns:
            pd.DataFrame: A DataFrame containing the results of the SQL query.
            None: If there is no connection to the database.

        Raises:
            RuntimeError: If there is no connection to the database or if an error occurs during query execution.
            ValidationError: If a module-specific error occurs during query execution.
        """

        raise NotImplementedError("You need to connect to a database first.")

    def get_dbschema(self):
        """
        Retrieve the database schema information.

        Returns:
            list: A list of dictionaries containing column information for the specified schemas.
        """
        raise NotImplementedError("You need to connect to a database first.")

    def get_schema_names(self):
        """
        Retrieve the names of all schemas in the database, excluding system schemas.

        This method executes a SQL query to fetch all schema names from the database.
        It filters out the 'information_schema' and any system schemas.
        If no schemas are found after filtering, a ValidationError is raised.

        Returns:
            list: A list of schema names present in the database, excluding system schemas.

        Raises:
            ValidationError: If no schemas are found in the database.
        """
        raise NotImplementedError("You need to connect to a database first.")

    @classmethod
    def register_connector(cls, db_type: str, connector):
        """
        Add a database connector class to the list of supported connectors.

        Args:
            db_type (str): The type of the database for which the connector is being registered.
            connector (class): The database connector class to be registered.

        Example:
            class MySQLConnector(DatabaseConnector):
                pass

            DatabaseConnector.register_connector("mysql", MySQLConnector)
        """
        if not hasattr(cls, "connectors"):
            cls.connectors = {}
        cls.connectors[db_type] = connector

    @staticmethod
    def get_connector(db_type: SupportedDBs):
        """
        Get the appropriate database connector class based on the provided database type.

        Args:
            db_type (SupportedDBs): The type of the database for which the connector is needed.

        Returns:
            class: The connector class corresponding to the specified database type.

        Raises:
            ValueError: If no connectors are registered or if the provided database type is not supported.
        """
        if not hasattr(DatabaseConnector, "connectors"):
            raise ValueError("No connectors registered.")
        if db_type in DatabaseConnector.connectors:
            return DatabaseConnector.connectors[db_type]
        raise ValueError(f"Unsupported database type: {db_type}")

    def get_default_training_plan(self):
        """
        Generate a default training plan based on the database schema, to be added to a vector store

        This method inspects the database schema to identify the relevant columns for databases, schemas, tables, and columns.
        It then constructs a training plan that includes documentation for each table in each database, detailing the columns
        present in the table and providing a preview of the first five rows of the table.

        Returns:
            TrainingPlan: An object containing the training plan with documentation for each table in the database.
        """
        db_schema = self.get_dbschema()
        # For each of the following, we look at the df columns to see if there's a match:
        database_column = db_schema.columns[
            db_schema.columns.str.lower().str.contains("database")
            | db_schema.columns.str.lower().str.contains("table_catalog")
        ].to_list()[0]
        database_list = db_schema[database_column].unique().tolist()
        schema_column = db_schema.columns[
            db_schema.columns.str.lower().str.contains("table_schema")
        ].to_list()[0]
        schema_list = db_schema[schema_column].unique().tolist()
        table_column = db_schema.columns[
            db_schema.columns.str.lower().str.contains("table_name")
        ].to_list()[0]
        table_list = db_schema[table_column].unique().tolist()
        column_column = db_schema.columns[
            db_schema.columns.str.lower().str.contains("column_name")
        ].to_list()[0]
        data_type_column = db_schema.columns[
            db_schema.columns.str.lower().str.contains("data_type")
        ].to_list()[0]

        plan = TrainingPlan([])
        for database in database_list:
            for table_schema in schema_list:
                for table in table_list:
                    df_columns_filtered_to_table = db_schema.query(
                        f'{database_column} == "{database}" and {schema_column} == "{table_schema}" and {table_column} == "{table}"'
                    )
                    doc = f"The following columns are in the {table} table in the {database} database:\n\n"
                    doc += df_columns_filtered_to_table[
                        [
                            database_column,
                            schema_column,
                            table_column,
                            column_column,
                            data_type_column,
                        ]
                    ].to_markdown()
                    plan._plan.append(
                        TrainingPlanItem(
                            item_type=TrainingPlanItem.ITEM_TYPE_IS,
                            item_group=f"{database}.{table_schema}",
                            item_name=table,
                            item_value=doc,
                        )
                    )

                    doc = f"Following are the first five rows of the {table} table in the {database} database:\n\n"
                    doc += (
                        self.run_sql(
                            f"SELECT * FROM {database}.{table_schema}.{table} LIMIT 5"
                        )
                        .fillna("nan")
                        .to_markdown(
                            floatfmt=".2f",
                            index=False,
                        )
                    )
                    plan._plan.append(
                        TrainingPlanItem(
                            item_type=TrainingPlanItem.ITEM_TYPE_IS,
                            item_group=f"{database}.{table_schema}",
                            item_name=table,
                            item_value=doc,
                        )
                    )
        return plan


@register(name="postgres")
class PostgresConnector(DatabaseConnector):
    """
    A class to manage connections and operations with a PostgreSQL database.

    Attributes:
        host (str): The hostname of the PostgreSQL server.
        port (Union[int, str]): The port number on which the PostgreSQL server is listening.
        database (str): The name of the database to connect to.
        user (str): The username to authenticate with.
        password (str): The password to authenticate with.
        schema (Union[list, str], optional): The schema(s) to use. Defaults to None.
        tables (list[str], optional): The list of tables to use. Defaults to None.
        conn (psycopg2.extensions.connection): The connection object to the PostgreSQL database.

    Methods:
        fetch_dataframes_dict(schema: Union[str, list] = None, tables: list[str] = None) -> dict[pd.DataFrame]:
            Fetches data from the specified schema and tables and returns it as a dictionary of pandas DataFrames.

        get_schema_names() -> list:
            Retrieves the names of all schemas in the database, excluding system schemas.

        get_dbschema() -> pd.DataFrame:
            Retrieves the schema information for the specified schemas.

        run_sql(sql: str) -> Union[pd.DataFrame, None]:
            Executes a SQL query and returns the result as a pandas DataFrame.
    """

    def __init__(
        self,
        host: str,
        port: Union[int, str],
        database: str,
        user: str,
        password: str,
        schema: Union[list, str] = None,
        tables: list[str] = None,
    ):
        self.host = host or os.getenv("POSTGRES_HOST")
        self.port = port or os.getenv("POSTGRES_PORT")
        self.database = database or os.getenv("POSTGRES_DB")
        self.user = user or os.getenv("POSTGRES_USER")
        self.password = password or os.getenv("POSTGRES_PASSWORD")
        if not all([self.host, self.port, self.database, self.user, self.password]):
            raise ImproperlyConfigured(
                "Please provide all required environment variables for Postgres connection:"
                " POSTGRES_HOST, POSTGRES_PORT, POSTGRES_DB, POSTGRES_USER, POSTGRES_PASSWORD"
            )
        self.schema = schema or os.getenv("POSTGRES_SCHEMA") or None
        if isinstance(self.schema, str):
            self.schema = [self.schema]
        self.tables = tables or os.getenv("POSTGRES_TABLES") or None
        (self.psycopg2,) = import_modules({"psycopg2": required_modules["psycopg2"]})
        try:
            self.conn = self.psycopg2.connect(
                host=self.host,
                database=self.database,
                user=self.user,
                password=self.password,
                port=self.port,
            )
        except self.psycopg2.Error as e:
            self.conn = None
            raise ValidationError(
                f"Error occurred while connecting to PostgreSQL database: {e}"
            ) from e
        if self.schema is None:
            self.schema = self.get_schema_names()

    def fetch_dataframes_dict(
        self,
        schema: Union[str, list] = None,
        tables: list[str] = None,
    ) -> dict[pd.DataFrame]:
        schema = schema or self.schema or None
        if isinstance(schema, str):
            schema = [schema]
        if schema is None:
            schema = self.get_schema_names()
        tables = tables or self.tables
        if self.conn is None:
            raise RuntimeError("No connection to the database.")
        try:
            cursor = self.conn.cursor()
            as_is = self.psycopg2.extensions.AsIs

            # Fetch all table names in the schema
            if tables is None:
                query = "SELECT TABLE_SCHEMA, TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA in (%s);"
                schema_names = ", ".join([f"'{s}'" for s in schema])
                cursor.execute(query, (as_is(schema_names),))
                tables = cursor.fetchall()
                tables = list({f"'{table[0]}'.'{table[1]}'" for table in tables})

            # Fetch all table contents and store in a dictionary as pandas dataframes
            dataframes = {tablename: "SELECT * FROM %s;" for tablename in tables}
            for tablename, query in dataframes.items():
                cursor.execute(query, (as_is(tablename),))
                dataframes[tablename] = pd.DataFrame(
                    cursor.fetchall(), columns=[desc[0] for desc in cursor.description]
                )
            # return the dataframes and the connection object
            return dataframes
        except self.psycopg2.Error as e:
            self.conn.rollback()
            raise ValidationError(
                f"Error occurred while connecting to PostgreSQL database: {e}"
            ) from e
        except Exception as e:
            self.conn.rollback()
            raise RuntimeError(
                f"Error occurred while fetching data from PostgreSQL table: {e}"
            ) from e

    def get_schema_names(self):
        all_schema_names = self.run_sql(
            "SELECT table_schema FROM information_schema.columns;"
        ).table_schema.unique()
        schema_names = [
            schema
            for schema in all_schema_names
            if schema != "information_schema" and not schema.startswith("pg_")
        ]
        if len(schema_names) == 0:
            raise ValidationError("No schema found in the database.")
        return schema_names

    def get_dbschema(self):
        schema_names = ", ".join([f"'{s}'" for s in self.schema])
        return self.run_sql(
            f"SELECT * FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA in ({schema_names});"
        )

    def run_sql(self, sql: str) -> Union[pd.DataFrame, None]:
        if self.conn is None:
            raise RuntimeError("No connection to the database.")
        try:
            cs = self.conn.cursor()
            cs.execute(sql)
            results = cs.fetchall()
            # Create a pandas dataframe from the results
            df = pd.DataFrame(results, columns=[desc[0] for desc in cs.description])
            return df

        except self.psycopg2.Error as e:
            self.conn.rollback()
            raise ValidationError(
                f"Error occurred while connecting to PostgreSQL database: {e}"
            ) from e

        except Exception as e:
            self.conn.rollback()
            raise RuntimeError(
                f"Error occurred while executing query on PostgreSQL table: {e}"
            ) from e


@register(name="redshift")
class RedshiftConnector(DatabaseConnector):
    """
    A class to manage connections and operations with an Amazon Redshift database.

    Attributes:
        host (str): The hostname of the Redshift server.
        port (int): The port number on which the Redshift server is listening.
        database (str): The name of the database to connect to.
        user (str): The username to authenticate with.
        password (str): The password to authenticate with.
        schema (Union[list, str]): The schema(s) to use. Defaults to None.
        tables (list[str], optional): The list of tables to use. Defaults to None.
        conn (redshift_connector.Connection): The connection object to the Redshift database.

    Methods:
        fetch_dataframes_dict(schema: str = None, tables: list[str] = None) -> dict[pd.DataFrame]:
            Fetches data from the specified schema and tables and returns it as a dictionary of pandas DataFrames.

        get_schema_names() -> list:
            Retrieves the names of all schemas in the database, excluding system schemas.

        get_dbschema() -> pd.DataFrame:
            Retrieves the schema information for the specified schemas.

        run_sql(sql: str) -> Union[pd.DataFrame, None]:
            Executes a SQL query and returns the result as a pandas DataFrame.
    """

    def __init__(
        self,
        host: str,
        database: str,
        user: str,
        password: str,
        port: int,
        schema: str,
        tables: list[str] = None,
    ):
        self.host = host or os.getenv("REDSHIFT_HOST") or 5439
        self.port = port or os.getenv("REDSHIFT_PORT")
        self.database = database or os.getenv("REDSHIFT_DB")
        self.user = user or os.getenv("REDSHIFT_USER")
        self.password = password or os.getenv("REDSHIFT_PASSWORD")
        if not all([self.host, self.port, self.database, self.user, self.password]):
            raise ImproperlyConfigured(
                "Please provide all required environment variables for Redshift connection:"
                " REDSHIFT_HOST, REDSHIFT_PORT, REDSHIFT_DB, REDSHIFT_USER, REDSHIFT_PASSWORD"
            )
        self.schema = schema or os.getenv("REDSHIFT_SCHEMA") or None
        if isinstance(self.schema, str):
            self.schema = [self.schema]
        self.tables = tables or os.getenv("REDSHIFT_TABLES") or None
        (self.redshift_connector,) = import_modules(
            {"redshift_connector": required_modules["redshift_connector"]}
        )
        try:
            self.conn = self.redshift_connector.connect(
                host=self.host,
                database=self.database,
                user=self.user,
                password=self.password,
                port=self.port,
            )
        except self.redshift_connector.InterfaceError as e:
            self.conn = None
            raise ValidationError(
                f"Error occurred while connecting to Redshift database: {e}"
            ) from e

        if self.schema is None:
            self.schema = self.get_schema_names()

    def fetch_dataframes_dict(
        self,
        schema: str = None,
        tables: list[str] = None,
    ) -> dict[pd.DataFrame]:
        schema = schema or self.schema or ["public"]
        tables = tables or self.tables
        if self.conn is None:
            raise RuntimeError("No connection to the database.")
        try:
            cursor = self.conn.cursor()
            # Fetch all table names in the schema
            if tables is None:
                schema_names = ", ".join([f"'{s}'" for s in schema])
                query = f"SELECT schemaname, tablename FROM pg_table_def WHERE schemaname in ({schema_names});"
                cursor.execute(query)
                tables = cursor.fetchall()
                tables = list({f"'{table[0]}'.'{table[1]}'" for table in tables})

            # Fetch all table contents and store in a dictionary as pandas dataframes
            dataframes = {
                tablename: f"SELECT * FROM {tablename};" for tablename in tables
            }
            for tablename, query in dataframes.items():
                cursor.execute(query)
                dataframes[tablename] = pd.DataFrame(
                    cursor.fetchall(), columns=[desc[0] for desc in cursor.description]
                )
            return dataframes
        except self.redshift_connector.InterfaceError as e:
            self.conn.rollback()
            raise ValidationError(
                f"Error occurred while connecting to Redshift database: {e}"
            ) from e
        except Exception as e:
            self.conn.rollback()
            raise RuntimeError(
                f"Error occurred while fetching data from Redshift table: {e}"
            ) from e

    def get_schema_names(self):
        all_schema_names = self.run_sql(
            "SELECT table_schema FROM information_schema.columns;"
        ).table_schema.unique()
        schema_names = [
            schema
            for schema in all_schema_names
            if schema != "information_schema" and not schema.startswith("pg_")
        ]
        if len(schema_names) == 0:
            raise ValidationError("No schema found in the database.")
        return schema_names

    def get_dbschema(self):
        schema_names = ", ".join([f"'{s}'" for s in self.schema])
        return self.run_sql(
            f"SELECT * FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA in ({schema_names});"
        )

    def run_sql(self, sql: str) -> Union[pd.DataFrame, None]:
        if self.conn is None:
            raise RuntimeError("No connection to the database.")
        try:
            cursor = self.conn.cursor()
            cursor.execute(sql)
            results = cursor.fetchall()
            # Create a pandas dataframe from the results
            df = pd.DataFrame(results, columns=[desc[0] for desc in cursor.description])
            return df
        except self.redshift_connector.InterfaceError as e:
            self.conn.rollback()
            raise ValidationError(
                f"Error occurred while connecting to Redshift database: {e}"
            ) from e
        except Exception as e:
            self.conn.rollback()
            raise RuntimeError(
                f"Error occurred while executing query on Redshift table: {e}"
            ) from e


@register(name="sqlite")
class SQLiteConnector(DatabaseConnector):
    """
    A class to manage connections and operations with an SQLite database.

    Attributes:
        db_path (str): The file path to the SQLite database.
        conn (sqlite3.Connection): The connection object to the SQLite database.

    Methods:
        __init__(db_path: str = None):
            Initializes the SQLiteConnector with the given database path or environment variable.

        _download_db(url: str) -> Union[str, None]:
            Downloads the SQLite database file from the given URL if it does not exist locally.

        create_database(db_path: str, df_dict: dict[pd.DataFrame]):
            Creates an SQLite database at the specified path using the provided dictionary of pandas DataFrames.

        fetch_dataframes_dict(tables: list[str] = None) -> dict[pd.DataFrame]:
            Fetches data from the specified tables and returns it as a dictionary of pandas DataFrames.

        run_sql(sql: str) -> Union[pd.DataFrame, None]:
            Executes a SQL query and returns the result as a pandas DataFrame.

        get_default_training_plan() -> TrainingPlan:
            Generates a default training plan based on the schema and data of the SQLite database.
    """

    def __init__(self, db_path: str = None):
        self.db_path = db_path or os.getenv("SQLITE_DB_PATH")
        self.db_path = SQLiteConnector._download_db(self.db_path)
        if self.db_path is None:
            raise ImproperlyConfigured(
                "Please provide a valid path to the SQLite database file."
            )
        self.conn = None
        try:
            self.conn = sqlite3.connect(self.db_path, check_same_thread=False)
        except sqlite3.Error as e:
            raise ValidationError(
                f"Error occurred while connecting to SQLite database: {e}"
            ) from e

    @staticmethod
    def _download_db(url: str) -> Union[str, None]:
        """
        Downloads a database file from the given URL if it does not already exist locally.

        Args:
            url (str): The URL of the database file to download.

        Returns:
            Union[str, None]: The local file path of the downloaded database file if successful,
            otherwise the original URL if the download fails or the file already exists locally.
        """
        url = urlparse(url).path
        if os.path.exists(url):
            return url
        try:
            filepath = os.path.basename(url)
            response = requests.get(url)
            response.raise_for_status()  # Check that the request was successful
            with open(filepath, "wb") as f:
                f.write(response.content)
            return filepath
        except Exception:
            return url

    def create_database(self, db_path: str, df_dict: dict[pd.DataFrame]):
        """
        Creates an SQLite database at the specified path and populates it with tables from the provided DataFrame dictionary.

        Args:
            db_path (str): The file path where the SQLite database will be created. If the path is invalid
                or empty, an in-memory database will be used.
            df_dict (dict[pd.DataFrame]): A dictionary where keys are table names and values are pandas
                DataFrames to be stored in the database.

        Returns:
            sqlite3.Connection: A connection object to the created SQLite database.

        Raises:
            ValidationError: If an error occurs while creating the SQLite database.
            RuntimeError: If an error occurs while executing a query on the SQLite database.
        """
        self.db_path = urlparse(db_path).path or self.db_path or ":memory:"

        parent = Path(self.db_path).parent
        os.makedirs(parent, exist_ok=True)

        from lyzr.data_analyzr.utils import translate_string_name

        try:
            self.conn = sqlite3.connect(self.db_path, check_same_thread=False)
            for name, df in df_dict.items():
                df = df.rename(
                    columns={col: translate_string_name(col) for col in df.columns}
                )
                df.to_sql(name, con=self.conn, index=False, if_exists="replace")
            return self.conn
        except sqlite3.Error as e:
            raise ValidationError(
                f"Error occurred while creating SQLite database: {e}"
            ) from e
        except Exception as e:
            raise RuntimeError(
                f"Error occurred while executing query on SQLite database: {e}"
            ) from e

    def fetch_dataframes_dict(self, tables: list[str] = None) -> dict[pd.DataFrame]:
        if self.conn is None:
            raise RuntimeError("No connection to the database.")
        try:
            if tables is None:
                cursor = self.conn.cursor()
                cursor.execute("SELECT name FROM sqlite_schema WHERE type='table';")
                tables = cursor.fetchall()
                tables = [table[0] for table in tables]
            dataframes = {
                table: pd.read_sql(f"SELECT * FROM {table};", con=self.conn)
                for table in tables
            }
            return dataframes
        except sqlite3.Error as e:
            raise ValidationError(
                f"Error occurred while fetching data from SQLite database: {e}"
            ) from e
        except Exception as e:
            raise RuntimeError(
                f"Error occurred while executing query on SQLite database: {e}"
            ) from e

    def run_sql(self, sql: str) -> Union[pd.DataFrame, None]:
        if self.conn is None:
            raise RuntimeError("No connection to the database.")
        try:
            df = pd.read_sql(sql, con=self.conn)
            return df
        except sqlite3.Error as e:
            raise ValidationError(
                f"Error occurred while executing query on SQLite database: {e}"
            ) from e
        except Exception as e:
            raise RuntimeError(
                f"Error occurred while executing query on SQLite database: {e}"
            ) from e

    def get_default_training_plan(self):

        tables_schema = self.run_sql("SELECT * FROM sqlite_master WHERE type='table';")
        plan = TrainingPlan([])

        for table_idx in tables_schema.index:
            table_info = tables_schema.loc[table_idx]

            table_columns = self.run_sql(
                f"PRAGMA table_info('{table_info['tbl_name']}')"
            )
            doc = (
                f"The following columns are in the {table_info['tbl_name']} table:\n\n"
            )
            doc += table_columns[["name", "type"]].to_markdown()

            plan._plan.append(
                TrainingPlanItem(
                    item_type=TrainingPlanItem.ITEM_TYPE_IS,
                    item_group=table_info["name"],
                    item_name=table_info["tbl_name"],
                    item_value=doc,
                )
            )

            doc = f"Following are the first five rows of the {table_info['tbl_name']} table:\n\n"
            doc += (
                self.run_sql(f"SELECT * FROM '{table_info['tbl_name']}' LIMIT 5")
                .fillna("nan")
                .to_markdown(
                    floatfmt=".2f",
                    index=False,
                )
            )
            plan._plan.append(
                TrainingPlanItem(
                    item_type=TrainingPlanItem.ITEM_TYPE_IS,
                    item_group=f"{table_info['name']}.public",
                    item_name=table_info["tbl_name"],
                    item_value=doc,
                )
            )
        return plan



"""
Logging functions for the lyzr package.
"""

import os
import re
import sys
import logging
import logging.handlers


def set_logger(
    name: str = None,
    logfilename: str = None,
    log_level: str = "INFO",
    print_log: bool = False,
):
    """
    Sets up a logger with specified configurations.

    Args:
        name (str, optional): The name of the logger. Defaults to "lyzr".
        logfilename (str, optional): The filename for the log file. Defaults to "lyzr".
        log_level (str, optional): The logging level (e.g., "INFO", "DEBUG"). Defaults to "INFO".
        print_log (bool, optional): If True, logs will also be printed to the console. Defaults to False.

    Returns:
        logging.Logger: Configured logger instance.

    Raises:
        ValueError: If an invalid log level is provided.

    Notes:
        - The log file will be in CSV format.
        - The logger will use a rotating file handler with a maximum file size of 25MB and up to 25 backup files.
        - Environment variables "log_level" and "logfilepath" will be set to the specified log level and log file path, respectively.
    """
    name = "lyzr" if name is None else name
    logfilename = "lyzr" if logfilename is None else logfilename
    logfilename = os.path.relpath(os.path.splitext(logfilename)[0] + ".csv")
    if os.path.dirname(logfilename).strip() != "":
        os.makedirs(os.path.dirname(logfilename), exist_ok=True)
    logger = logging.getLogger(name=name)
    if logger.hasHandlers():
        for handler in logger.handlers:
            try:
                handler.close()
            except Exception:
                pass
        logger.handlers.clear()
    numeric_level = getattr(logging, log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError("Invalid log level: %s" % log_level)
    logger.setLevel(numeric_level)
    # write to console
    if print_log:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(
            logging.Formatter(
                get_console_log_format(),
                datefmt="%Y-%m-%d %H:%M:%S %Z",
                style="{",
            )
        )
        handler.setLevel(log_level)
        logger.addHandler(handler)
    # write to rotating csv files
    filehandler = logging.handlers.RotatingFileHandler(
        filename=logfilename,
        mode="a",
        maxBytes=26214400,  # maxBytes=25MB
        backupCount=25,
    )
    fileformatter = CustomFormatter(
        get_csv_log_format(),
        datefmt="%Y-%m-%d %H:%M:%S %Z",
        style="{",
    )
    filehandler.setLevel(logging.INFO)
    filehandler.setFormatter(fileformatter)
    logger.addHandler(filehandler)

    os.environ["log_level"] = str(log_level)
    os.environ["logfilepath"] = logfilename
    return logger


class CustomFormatter(logging.Formatter):
    """
    Custom logging formatter to handle dynamic fields and format log records.

    Methods:
        format(record: logging.LogRecord) -> str
            Formats the specified log record as text. Ensures that all dynamic fields
            specified in the format string are present in the log record, setting them
            to None if they are missing. Handles special formatting for the 'traceback'
            field and ensures that newlines are properly represented.
    """

    def format(self, record: logging.LogRecord) -> str:
        arg_pattern = re.compile(r"\{(\w+)\}")
        arg_names = [x.group(1) for x in arg_pattern.finditer(self._fmt)]
        for field in arg_names:
            if not hasattr(record, field):
                setattr(record, field, None)
            record_value = getattr(record, field)
            if (
                field == "traceback"
                and isinstance(record_value, str)
                and (
                    (record_value.strip()[0] != "[")
                    and (record_value.strip()[-1] != "]")
                )
            ):
                setattr(record, field, record_value.splitlines())
            setattr(record, field, str(record_value).replace("\n", "\\n"))
        record.msg = record.getMessage().strip()
        record.msg = record.getMessage().replace("\n", "\\n")
        return super().format(record)


def get_csv_log_format():
    """Returns the format string for logging records to CSV files."""
    columns = [
        "asctime",
        "name",
        "levelname",
        "filename",
        "funcName",
        "lineno",
        "message",
        "function",
        "traceback",
        "input_args",
        "input_kwargs",
        "response",
    ]
    sep = ",;"
    log_format = sep.join(["{" + x + "}" for x in columns])
    return log_format


def get_console_log_format():
    """Returns the format string for logging records to the console."""
    log_format = "{asctime} | {name} — {levelname} — {filename}: {funcName}: {lineno} — {message}"
    return log_format


def read_csv_log(filename: str):
    """
    Reads a CSV log file and returns it as a pandas DataFrame.

    Args:
        filename (str): The path to the CSV log file.

    Returns:
        pandas.DataFrame: A DataFrame containing the log data with predefined columns.

    Notes:
        - The expected columns in the CSV file are:
            - asctime
            - name
            - levelname
            - filename
            - funcName
            - lineno
            - message
            - function
            - traceback
            - input_args
            - input_kwargs
            - response
        - The CSV file is expected to use ",;" as the separator.
    """
    import pandas as pd

    columns = [
        "asctime",
        "name",
        "levelname",
        "filename",
        "funcName",
        "lineno",
        "message",
        "function",
        "traceback",
        "input_args",
        "input_kwargs",
        "response",
    ]
    sep = ",;"
    df = pd.read_csv(filename, sep=sep, names=columns, engine="python")
    return df



import os
from typing import Optional, Literal
from lyzr.base.llms import LLM, get_model
from lyzr.base.errors import MissingValueError


class Generator:
    def __init__(
        self,
        api_key: Optional[str] = None,
        model_type: Optional[Literal["openai"]] = None,
        model: Optional[LLM] = None,
    ):
        self.api_key = api_key or os.environ.get("OPENAI_API_KEY")
        if self.api_key is None:
            raise MissingValueError("API key")
        self.model = model or get_model(
            api_key=self.api_key,
            model_type=model_type or os.environ.get("MODEL_TYPE") or "openai",
        )  # change get_model in lyzr.base.llms to accept **kwargs

    def generate(
            self, 
            text:str,
            persona: Optional[str] = "Not Specified",
            instructions: Optional[str] = "Short Story" #Could be a Poem, a Children's story or even a tweet 
            ) -> str:
        '''
    Generates content in various styles such as a children's story, a poem, or even a tweet from the provided text using OpenAI's GPT-4 model. This function is designed to expand a byte sized prompt into a more elaborate format, according to the specified instructions.
    
    Parameters:
    
    - `text` (str): The substantial text or conversation input that needs to be elaborated or expanded.
    
    - `persona` (Optional[str], default = "Not Specified"): Specifies the persona or audience for which the content is tailored. This parameter helps in customizing the tone and style of the generated content to better resonate with the intended audience.
    
    - `instructions` (Optional[str], default = "Short Story"): Specifies the type of output desired. Options include "Mini Essay" for a detailed narrative, "Poem" for poetic content, "Children's Story" for content suitable for children, or "Tweet" for extremely concise content suitable for social media platforms like Twitter. This parameter influences the instruction set given to the AI, tailoring its approach to content generation.
    
    Return:
    
    - A string containing the generated content that effectively expands the essence of the original text into the desired format. This output aims to retain all pertinent information and key themes while presenting them in a clear, coherent, and stylistically appropriate manner.
    
    Example Usage:
    
    ```python
    from lyzr import Generator
    
    # Initialize the content generator
    generator = Generator(api_key="your_openai_api_key")
    
    # Provide the text to be condensed and specify the desired style
    text = "Prompt or idea that you want to expand upon"
    story = generator.generate(text, instructions='story')
    print(story)
    
    # You can also specify the persona for which the content is tailored
    persona = "Tech Enthusiasts"
    condensed_content = generator.generate(text, persona=persona, instructions='Tweet')
    print(condensed_content)
    ```
    
    This functionality leverages advanced language model capabilities for creating concise and accurate representations of larger bodies of text, adjustable to various output styles and tailored to specific personas for enhanced utility in information processing and content creation scenarios.
        '''
        if self.model.model_name != "gpt-4":
            if self.model.model_type == "openai":
                self.model = get_model(
                    api_key=self.api_key,
                    model_type=self.model.model_type,
                    model_name="gpt-4",
                )
            else:
                raise ValueError(
                    "This function only works with the OpenAI's 'gpt-4' model."
                )

        # The system message acts as the prompt for the AI.
        system_message = f'''You are an Expert CONTENT CREATOR. Your task is to DEVELOP a VARIETY of TEXT-BASED CONTENT that could range from BLOGS to TWEETS.

The target audience of the content: {persona}

The Format of the content should be based on these instructions: {instructions}  
        
Here's how you can approach this task:

1. IDENTIFY the target audience for whom you will be creating content. Understand their interests, needs, and preferences.
2. CHOOSE the type of content you wish to create first—whether it's a blog post, tweet, article, or any other form of written communication.
3. DECIDE on the topics that will RESONATE with your audience and align with your content strategy or goals.
4. DRAFT an outline or key points for each piece of content to ensure STRUCTURE and FLOW.
5. WRITE the initial draft focusing on ENGAGEMENT and VALUE delivery for your readers or followers.
6. EMPLOY a conversational tone or formal style according to the platform and type of content you are creating.
7. EDIT and REVISE your work to enhance CLARITY, GRAMMAR, and COHERENCE before publishing.
8. DON'T display anything that is not relevant to the content such as comments or instructions.

Now Take a Deep Breath.'''

        # Format the user's message that will be sent to the model.
        user_message = text
        self.model.set_messages(
            model_prompts=[
                {"role": "system", "text": system_message},
                {"role": "user", "text": user_message},
            ]
        )
        # Use the LLM instance to communicate with OpenAI's API.
        response = self.model.run()

        # Parse the response to extract the notes.
        notes = response.choices[0].message.content

        return notes


DATA_ANALYZR_PROMPTS = {
    "ai_queries": {
        "system": {
            "context": "You are an Expert DATA ANALYST with extensive knowledge of PANDAS and SCIKIT-LEARN libraries. Your task is to CREATE 20 NATURAL LANGUAGE QUERIES to analyze the provided dataset(s) using these libraries.\n\n",
            "external_context": "{context}",
            "task": "You MUST ensure that each query is simple, easy to understand, and can be answered with the data in the dataset.\n\nFollow this step-by-step guide:\n1. UNDERSTAND THE DATASET: Review the sample dataset provided to grasp its structure, variables, and context.\n2. GENERATE QUERIES: Create natural language queries for each applicable category: Exploratory Analysis, Regression Analysis, Correlation Analysis, Clustering Analysis, and Time Series Analysis.\n3. CATEGORIZE: Only use categories that are applicable to the given dataset. Skip categories which are not applicable.\n4. ENSURE RELEVANCE: Make sure every query can be answered with the given dataset. Do NOT include hypothetical or non-applicable queries.\n5. FORMAT IN JSON: Organize your queries into a JSON format as shown below:\n```json\n{schema}\n```\n\n",
            "closing": "Remember, I’m going to tip $300K for a BETTER SOLUTION!\n\nNow Take a Deep Breath.",
        },
        "user": {"inputs": "{df_details}"},
    },
    "analysis_code": {
        "system": {
            "context": "You are an Expert DATA ANALYST and PYTHON CODER. Your task is to RESPOND with precise Python code based on the questions provided by the user.\n\n",
            "external_context": "{context}",
            "task": "Please follow these steps:\n1. READ the user's question CAREFULLY to understand what Python code is being requested.\n2. WRITE the Python code that directly answers the user's question.\n3. ENSURE that your response contains ONLY the Python code without any additional explanations or comments.\n4. VERIFY that your Python code is SYNTACTICALLY CORRECT and adheres to standard Pythonic practices.\n5. You code must SAVE the result to `result`.\n6. Whenever possible your code should OUTPUT a pandas dataframe.\n7. You may use triple backticks ``` before and after the code block.\n8. Do NOT add comments your code.\n\n",
            "closing": "You MUST provide clean and efficient Python code as a response, and remember, I'm going to tip $300K for a BETTER SOLUTION!\n\nNow Take a Deep Breath.\n\n",
            "doc_addition_text": "You may use the following documentation to understand the schema of the data:\n{doc}\n",
            "history": "Also use responses to past questions to guide you.\n\n",
            "locals": "The following local environment variables are available to you:\n{locals}\n\n",
        }
    },
    "sql_question_gen": {
        "system": {
            "task": "The user will give you SQL and you will try to guess what the business question this query is answering.\nReturn just the question without any additional explanation.\nDo not reference the table name in the question."
        }
    },
    "txt_to_sql": {
        "system": {
            "context": "You are an Expert SQL CODER. Your task is to RESPOND with precise SQL queries based on the questions provided by the user.\n\n",
            "external_context": "{context}",
            "task": "Please follow these steps:\n1. READ the user's question CAREFULLY to understand what SQL query is being requested.\n2. WRITE the SQL code that directly answers the user's question.\n3. ENSURE that your response contains ONLY the SQL code without any additional explanations or comments.\n4. VERIFY that your SQL code is SYNTACTICALLY CORRECT and adheres to standard SQL practices.\n5. You may use triple backticks ``` before and after the code block.\n\nYou MUST provide clean and efficient SQL queries as a response, and remember, I'm going to tip $300K for a BETTER SOLUTION!\n\nNow Take a Deep Breath.\n\n",
            "ddl_addition_text": "You may use the following DDL statements as a reference for what tables might be available:\n{ddl}\n",
            "doc_addition_text": "You may use the following documentation to understand the schema of the tables:\n{doc}\n",
            "history": "Also use responses to past questions to guide you.\n\n",
        }
    },
    "insights": {
        "system": {
            "context": "You are an Expert DATA ANALYST and COMMUNICATOR. Your task is to INTERPRET complex analytics results and TRANSLATE them into SIMPLE, UNDERSTANDABLE insights for business users and data analysts.\n\n",
            "external_context": "{context}",
            "task": "Proceed with the following steps:\n\n1. ANALYZE the user query, the analysis code, and the analysis output to fully comprehend the results derived from the initial dataset.\n2. SIMPLIFY the findings by creating clear explanations that resonate with both business users and data analysts, ensuring that you use plain language.\n3. ACCURATELY ROUND all relevant numbers to TWO DECIMAL PLACES to complement the analysis output.\n4. RANK your insights based on their significance and SHARE only the top {n_insights}.\n5. FORMAT these insights as BULLET POINTS for clarity and succinctness.\n\nYou MUST adhere to these guidelines:\n\n- Present ONLY THE LIST of insights without titles or additional information.\n- Ensure that each insight is DIRECTLY TIED to a corresponding data point from the analysis output.\n\nI’m going to tip $300K for a BETTER SOLUTION!\n\nTake a Deep Breath.",
        },
        "user": {
            "inputs": "Today is {date}.\nuser query: {user_input}\nanalysis code:\n{analysis_code}\n\nanalysis output:\n{analysis_output}"
        },
    },
    "recommendations": {
        "system": {
            "context": "You are an Expert STRATEGIC ADVISOR. Your task is to FORMULATE TOP-NOTCH RECOMMENDATIONS.\n\n",
            "external_context": "{context}",
            "task_no_insights": "You have been given the original query asked by the user and the analysis results generated.\nDevelop recommendations that the user can implement to improve their intended outcome.\nGive numbers and exact values to support your recommendations.\n\n",
            "task_with_insights": "You have been given the original query asked by the user, the analysis results generated, and the analysis insights.\nDevelop recommendations that the user can implement to improve their intended outcome.\nGive numbers and exact values to support your recommendations.\n\n",
            "json_type": "Rank all your recommendations and only share the top {n_recommendations} ones. For each of the {n_recommendations} recommendations, provide {output_json_keys}.\nPresent the output as a JSON object, with the following schema:\n{json_schema}\n\n",
            "text_type": "Rank all your recommendations and only share the top {n_recommendations} ones. Share your recommendations as a numbered list.\n\n",
            "closing": "Focus on clarity and succinctness.\n\nDo not describe the dataset or the prompt.\nDo not speak about charts.\nDo not share any title.",
        },
        "user": {"inputs": "{user_input}\n{insights}\n{analysis_output}"},
    },
    "tasks": {
        "system": {
            "context": "You are the world's best to-do creator and assigner. All the to-dos are executable in 60 minutes or less.\nYou can understand any recommendation or set of recommendations and break it down into a list of to-do's that the user could do in 60 minutes or less.\nThese to-dos will help the user execute the larger recommendations.\n\n",
            "external_context": "{context}",
            "task": "Generate a list of to-dos that are not complex and don't require multiple tasks to be executed at once.\nThey should be sequential and the user should be able to complete the to-dos one at a time.\n\nGenerate to-dos that are specific and quantifiable, ensuring that each task mentions a clear numeric target.\nEach task should be feasible to complete within a 2 to 3-hour timeframe.\nFor example, instead of saying 'Speak to customers', the task should be 'Speak to 10 customers'.\nLikewise, instead of 'Create a list of documents', it should specify 'Create a list of 30 documents'.\n\n",
            "closing": "Use the following format:\n{tasks_format}\n\nEach to-do should be between 15 and 25 words. And generate no more than {n_tasks} to-dos. \n\nThe to-dos should always start with one of the below:\nCreate a list,\nWrite down, \nSpeak to,\nSetup 30 minutes to dive deep and analyze,\nPlan to,\nDo,\nReview,\nDraft,\nComplete,\nBegin,\nDiscuss,\nSchedule,\nConfirm,\nReach out to,\nTest,\nAttend,\nAllocate time for",
        },
        "user": {
            "inputs": "User Query: {user_input}\n\nAnalysis results & insights: {insights}\n\nRecommendations: {recommendations}"
        },
    },
    "plotting_code": {
        "system": {
            "context": "You are an Expert DATA ANALYST and PYTHON CODER. Your task is to RESPOND with precise Python code based on the questions provided by the user.\n\n",
            "external_context": "{context}",
            "sql_plot": "Please follow these steps:\n1. READ the user's question CAREFULLY.\n2. UNDERSTAND what plot can be generated to answer the question.\n3. If needed, USE the 'conn' object to query the database with `pd.read_sql('SQL query here', conn.conn)`.\n4. WRITE the Python code that makes a figure `fig` with this plot.\n5. ENSURE that your response contains ONLY the code without any additional explanations or comments.\n4. VERIFY that your code is SYNTACTICALLY CORRECT and adheres to standard practices.\n5. You code must SAVE THE PLOT to `fig`.\n6. You may use triple backticks ``` before and after the code block.\n7. Do NOT add comments to your code.\n\nYou MUST provide clean and efficient code as a response, and remember, I'm going to tip $300K for a BETTER SOLUTION!\n\nNow Take a Deep Breath.\n\n",
            "python_plot": "Please follow these steps:\n1. READ the user's question CAREFULLY.\n2. UNDERSTAND what plot can be generated to answer the question.\n3. WRITE the Python code that makes a figure `fig` with this plot.\n3. ENSURE that your response contains ONLY the Python code without any additional explanations or comments.\n4. VERIFY that your Python code is SYNTACTICALLY CORRECT and adheres to standard Pythonic practices.\n5. You code must SAVE THE PLOT to `fig`.\n6. You may use triple backticks ``` before and after the code block.\n7. Do NOT add comments to your code.\n\nYou MUST provide clean and efficient Python code as a response, and remember, I'm going to tip $300K for a BETTER SOLUTION!\n\nNow Take a Deep Breath.\n\n",
            "doc_addition_text": "You may use the following documentation to understand the schema of the {db_type}:\n{doc}\n",
            "python_examples_text": "You may use the following examples to guide you:\n{python_examples}\n",
            "sql_examples_text": "You may use the following examples to guide you:\n{sql_examples}\n",
            "history": "Also use responses to past questions to guide you.",
            "locals": "The following local environment variables are available to you:\n{locals}\n",
        }
    },
}



from typing import Union, Optional, List

from llama_index import ServiceContext, VectorStoreIndex
from llama_index.embeddings.utils import EmbedType
from llama_index.indices.query.base import BaseQueryEngine

from lyzr.utils.search_utils import (
    add_pdf_,
    add_docx_,
    add_text_,
    add_webpage_,
    add_website_,
    add_youtube_
)


class SearchAgent:
    def __init__(self) -> None:
        return None

    @staticmethod
    def from_instances(
        vector_store_index: VectorStoreIndex, service_context: ServiceContext, **kwargs
    ) -> BaseQueryEngine:
        return vector_store_index.as_query_engine(
            service_context=service_context, **kwargs
        )

    @staticmethod
    def add_pdf(
        input_dir: Optional[str] = None,
        input_files: Optional[List] = None,
        exclude_hidden: bool = True,
        filename_as_id: bool = True,
        recursive: bool = True,
        required_exts: Optional[List[str]] = None,
        system_prompt: str = None,
        query_wrapper_prompt: str = None,
        embed_model: Union[str, EmbedType] = "default",
        llm_params: dict = None,
        vector_store_params: dict = None,
        service_context_params: dict = None,
        query_engine_params: dict = None,
        retriever_params: dict = None,
    ) -> BaseQueryEngine:
        return add_pdf_(
            input_dir=input_dir,
            input_files=input_files,
            exclude_hidden=exclude_hidden,
            filename_as_id=filename_as_id,
            recursive=recursive,
            required_exts=required_exts,
            system_prompt=system_prompt,
            query_wrapper_prompt=query_wrapper_prompt,
            embed_model=embed_model,
            llm_params=llm_params,
            vector_store_params=vector_store_params,
            service_context_params=service_context_params,
            query_engine_params=query_engine_params,
            retriever_params=retriever_params,
        )

    @staticmethod
    def add_docx(
        input_dir: Optional[str] = None,
        input_files: Optional[List] = None,
        exclude_hidden: bool = True,
        filename_as_id: bool = True,
        recursive: bool = True,
        required_exts: Optional[List[str]] = None,
        system_prompt: str = None,
        query_wrapper_prompt: str = None,
        embed_model: Union[str, EmbedType] = "default",
        llm_params: dict = None,
        vector_store_params: dict = None,
        service_context_params: dict = None,
        query_engine_params: dict = None,
        retriever_params: dict = None,
    ) -> BaseQueryEngine:
        return add_docx_(
            input_dir=input_dir,
            input_files=input_files,
            exclude_hidden=exclude_hidden,
            filename_as_id=filename_as_id,
            recursive=recursive,
            required_exts=required_exts,
            system_prompt=system_prompt,
            query_wrapper_prompt=query_wrapper_prompt,
            embed_model=embed_model,
            llm_params=llm_params,
            vector_store_params=vector_store_params,
            service_context_params=service_context_params,
            query_engine_params=query_engine_params,
            retriever_params=retriever_params,
        )

    @staticmethod
    def add_text(
        input_dir: Optional[str] = None,
        input_files: Optional[List] = None,
        exclude_hidden: bool = True,
        filename_as_id: bool = True,
        recursive: bool = True,
        required_exts: Optional[List[str]] = None,
        system_prompt: str = None,
        query_wrapper_prompt: str = None,
        embed_model: Union[str, EmbedType] = "default",
        llm_params: dict = None,
        vector_store_params: dict = None,
        service_context_params: dict = None,
        query_engine_params: dict = None,
        retriever_params: dict = None,
    ) -> BaseQueryEngine:
        return add_text_(
            input_dir=input_dir,
            input_files=input_files,
            exclude_hidden=exclude_hidden,
            filename_as_id=filename_as_id,
            recursive=recursive,
            required_exts=required_exts,
            system_prompt=system_prompt,
            query_wrapper_prompt=query_wrapper_prompt,
            embed_model=embed_model,
            llm_params=llm_params,
            vector_store_params=vector_store_params,
            service_context_params=service_context_params,
            query_engine_params=query_engine_params,
            retriever_params=retriever_params,
        )

    @staticmethod
    def add_webpage(
        url: Optional[str] = None,
        system_prompt: str = None,
        query_wrapper_prompt: str = None,
        embed_model: Union[str, EmbedType] = "default",
        llm_params: dict = None,
        vector_store_params: dict = None,
        service_context_params: dict = None,
        query_engine_params: dict = None,
        retriever_params: dict = None,
    ) -> BaseQueryEngine:
        return add_webpage_(
            url=url,
            system_prompt=system_prompt,
            query_wrapper_prompt=query_wrapper_prompt,
            embed_model=embed_model,
            llm_params=llm_params,
            vector_store_params=vector_store_params,
            service_context_params=service_context_params,
            query_engine_params=query_engine_params,
            retriever_params=retriever_params,
        )

    @staticmethod
    def add_website(
        url: Optional[str] = None,
        system_prompt: str = None,
        query_wrapper_prompt: str = None,
        embed_model: Union[str, EmbedType] = "default",
        llm_params: dict = None,
        vector_store_params: dict = None,
        service_context_params: dict = None,
        query_engine_params: dict = None,
        retriever_params: dict = None,
    ) -> BaseQueryEngine:
        return add_website_(
            url=url,
            system_prompt=system_prompt,
            query_wrapper_prompt=query_wrapper_prompt,
            embed_model=embed_model,
            llm_params=llm_params,
            vector_store_params=vector_store_params,
            service_context_params=service_context_params,
            query_engine_params=query_engine_params,
            retriever_params=retriever_params,
        )

    @staticmethod
    def add_youtube(
        urls: List[str] = None,
        system_prompt: str = None,
        query_wrapper_prompt: str = None,
        embed_model: Union[str, EmbedType] = "default",
        llm_params: dict = None,
        vector_store_params: dict = None,
        service_context_params: dict = None,
        query_engine_params: dict = None,
        retriever_params: dict = None,
    ) -> BaseQueryEngine:
        return add_youtube_(
            urls=urls,
            system_prompt=system_prompt,
            query_wrapper_prompt=query_wrapper_prompt,
            embed_model=embed_model,
            llm_params=llm_params,
            vector_store_params=vector_store_params,
            service_context_params=service_context_params,
            query_engine_params=query_engine_params,
            retriever_params=retriever_params,
        )



from lyzr.data_analyzr.analysis_handler.sql import TxttoSQLFactory
from lyzr.data_analyzr.analysis_handler.pythonic import PythonicAnalysisFactory
from lyzr.data_analyzr.analysis_handler.plotter import PlotFactory



"""
Vector store for storing and querying database information, past queries and their associated code.
"""

# standard library imports
import os
import json
import logging

# local imports
from lyzr.base.errors import (
    ValidationError,
    MissingValueError,
    DependencyError,
)

vector_store_collections = ["sql", "ddl", "documentation", "python", "plot"]
logging.getLogger("chromadb").setLevel(logging.CRITICAL)


class ChromaDBVectorStore:
    """
    A class to manage a vector store using ChromaDB for storing and querying database
    information, DDL statements, past queries and their associated code - SQL, pythonic and plotting code.

    Attributes:
        chroma_client (chromadb.PersistentClient): ChromaDB client instance.
        sql_collection (chromadb.Collection): Collection for storing SQL queries.
        documentation_collection (chromadb.Collection): Collection for storing documentation.
        ddl_collection (chromadb.Collection): Collection for storing DDL statements.
        python_collection (chromadb.Collection): Collection for storing Python code.
        plot_collection (chromadb.Collection): Collection for storing plotting code.
        logger (logging.Logger): Logger instance for logging information and errors.

    Methods:
        __init__(path, remake_store, training_plan, logger):
            Initializes a ChromaDBVectorStore instance.

        make_chroma_client(path):
            Initializes the ChromaDB client and sets up collections for various data types.

        remake_vector_store():
            Remakes the vector store by iterating through existing collections and recreating them if necessary.

        add_training_data(question, sql, ddl, plot_code, python_code, documentation, plan):
            Adds a training plan or individual components to the ChromaDBVectorStore.

        remake_collection(collection_name):
            Remakes a specified collection in the vector store.

        _extract_documents(query_results):
            Extracts and processes documents from the query results.

        get_related_ddl(question):
            Retrieves related DDL queries based on the question.

        get_related_documentation(question):
            Retrieves related documentation based on the question.

        get_related_sql_queries(user_input):
            Retrieves similar question-SQL query pairs based on user input.

        get_related_python_code(question):
            Retrieves similar question-Python code pairs based on the question.

        get_related_plotting_code(question):
            Retrieves similar question-plotting code pairs based on the question.

        generate_embedding(data):
            Generates an embedding for the given data.

        add_question_sql(question, sql):
            Adds a question and its corresponding SQL query to the collection.

        add_ddl(ddl):
            Adds a DDL statement to the collection.

        add_documentation(documentation):
            Adds documentation to the collection.

        add_python_code(question, python_code):
            Adds a question and its corresponding Python code to the collection.

        add_plot_code(question, plot_code):
            Adds a question and its corresponding plotting code to the collection.
    """

    from lyzr.data_analyzr.utils import deterministic_uuid
    from lyzr.data_analyzr.db_connector import (
        TrainingPlanItem,
        DatabaseConnector,
        DatabaseConnector,
        TrainingPlanItem,
        DatabaseConnector,
        TrainingPlanItem,
        TrainingPlan,
    )

    def __init__(
        self,
        path: str = None,  # Path to the directory where the ChromaDB data is stored
        remake_store: bool = False,  # If True, the store will be recreated
        training_plan_func: callable = None,  # Function to generate a training plan
        logger: logging.Logger = None,
        **kwargs,
    ):
        """
        Initialize a ChromaDBVectorStore at the specified path.

        Args:
            path (str, optional): Path to the directory where the ChromaDB data is stored. If not provided,
                a MissingValueError is raised.
            remake_store (bool, optional): If True, the store will be recreated. Defaults to False.
            training_plan_func (callable, optional): Function to generate a training plan. Defaults to None. Must return a TrainingPlan object.
            logger (logging.Logger, optional): Logger instance for logging information and errors. Defaults to None.

        Raises:
            MissingValueError: If the 'path' argument is not provided.

        Procedure:
        - If the specified path exists:
            - Initializes the ChromaDB client with the given path.
            - If remake_store is True, attempts to recreate the vector store and logs the process.
            - If a training_plan_func is provided:
                - Calls the function to generate a training plan.
                - Adds this training plan to the vector store.
        - If the specified path does not exist:
            - Creates the directory at the specified path.
            - Initializes the ChromaDB client with the given path.
            - If a training_plan_func is provided:
                - Calls the function to generate a training plan.
                - Adds this training plan to the vector store.

        Example:
            vector_store = ChromaDBVectorStore(
                path="path/to/vector_store",
                remake_store=False,
                training_plan_func=training_plan_func,
                logger=logger,
            )
            docs = vector_store.get_related_documentation(question)
            example_sql = vector_store.get_similar_question_sql(question)
            example_python_code = vector_store.get_similar_python_code(question)
            example_plot_code = vector_store.get_similar_plotting_code(question)
        """
        if path is None:
            raise MissingValueError("path")

        if os.path.exists(path):
            self.make_chroma_client(path=path)
            if remake_store:
                logger.info(f"Remaking vector store at {path}.")
                if not self.remake_vector_store():
                    logger.error(f"Failed to recreate vector store at {path}.")
                if training_plan_func is not None:
                    self.add_training_data(plan=training_plan_func(**kwargs))
            else:
                logger.info(f"Vector store exists at {path}. Using existing store.")
        else:
            os.makedirs(path)
            logger.info(f"Creating vector store at {path}.")
            self.make_chroma_client(path=path)
            if training_plan_func is not None:
                self.add_training_data(plan=training_plan_func(**kwargs))

    def make_chroma_client(self, path: str):
        """
        Initializes the ChromaDB client and sets up collections for various data types.

        Args:
            path (str): The file path where the ChromaDB client will store its data.

        Raises:
            DependencyError: If the required ChromaDB module is not installed.
        """
        try:
            import chromadb
            from chromadb.config import Settings
            from chromadb.utils import embedding_functions
        except ImportError:
            raise DependencyError(
                {
                    "chromadb": "chromadb==0.4.22",
                }
            )

        self.embedding_function = embedding_functions.DefaultEmbeddingFunction()

        self.chroma_client = chromadb.PersistentClient(
            path=path, settings=Settings(anonymized_telemetry=False)
        )
        self.sql_collection = self.chroma_client.get_or_create_collection(
            name="sql", embedding_function=self.embedding_function
        )
        self.documentation_collection = self.chroma_client.get_or_create_collection(
            name="documentation", embedding_function=self.embedding_function
        )
        self.ddl_collection = self.chroma_client.get_or_create_collection(
            name="ddl", embedding_function=self.embedding_function
        )
        self.python_collection = self.chroma_client.get_or_create_collection(
            name="python", embedding_function=self.embedding_function
        )
        self.plot_collection = self.chroma_client.get_or_create_collection(
            name="plot", embedding_function=self.embedding_function
        )

    def remake_vector_store(self):
        """
        Remakes the vector store by iterating through existing collections and
        recreating them if necessary. Also ensures that all required collections
        are present in the vector store.

        Returns:
            bool: True if all collections were successfully remade, False otherwise.
        """
        remake = True
        all_collections = vector_store_collections
        existing_collections = self.chroma_client.list_collections()
        for col in existing_collections:
            remake = remake and self.remake_collection(collection_name=col.name)
        for col in all_collections:
            if col in existing_collections:
                continue
            setattr(
                self,
                f"{col}_collection",
                (
                    self.chroma_client.get_or_create_collection(
                        name=col, embedding_function=self.embedding_function
                    )
                ),
            )
        return remake

    def add_training_data(
        self,
        question: str = None,
        sql: str = None,
        ddl: str = None,
        plot_code: str = None,
        python_code: str = None,
        documentation: str = None,
        plan: TrainingPlan = None,
    ) -> str:
        """
        Add a training plan or individual components to the ChromaDBVectorStore.

        Args:
            question (str, optional): The question to be added.
            sql (str, optional): The SQL query associated with the question.
            ddl (str, optional): The Data Definition Language (DDL) statement to be added.
            plot_code (str, optional): The code for generating a plot associated with the question.
            python_code (str, optional): The Python code associated with the question.
            documentation (str, optional): The documentation to be added.
            plan (TrainingPlan, optional): A TrainingPlan object containing multiple items to be added.

        Returns:
            str: A confirmation message or the result of adding the documentation.

        Raises:
            ValidationError: If a question is provided without an associated answer (SQL, plot code, or Python code).
        """
        if (
            (question and not sql)
            and (question and not plot_code)
            and (question and not python_code)
        ):
            raise ValidationError("Please also provide an answer to the question.")

        if documentation:
            return self.add_documentation(documentation)

        if sql:
            return self.add_question_sql(question=question, sql=sql)

        if ddl:
            return self.add_ddl(ddl)

        if plot_code:
            self.add_plot_code(question=question, plot_code=plot_code)

        if python_code:
            self.add_python_code(question=question, python_code=python_code)

        if plan:
            for item in plan._plan:
                if item.item_type == ChromaDBVectorStore.TrainingPlanItem.ITEM_TYPE_DDL:
                    self.add_ddl(item.item_value)
                elif (
                    item.item_type == ChromaDBVectorStore.TrainingPlanItem.ITEM_TYPE_IS
                ):
                    self.add_documentation(item.item_value)
                elif (
                    item.item_type == ChromaDBVectorStore.TrainingPlanItem.ITEM_TYPE_SQL
                ):
                    self.add_question_sql(question=item.item_name, sql=item.item_value)
                elif (
                    item.item_type == ChromaDBVectorStore.TrainingPlanItem.ITEM_TYPE_PY
                ):
                    self.add_python_code(
                        question=item.item_name, python_code=item.item_value
                    )
                elif (
                    item.item_type
                    == ChromaDBVectorStore.TrainingPlanItem.ITEM_TYPE_PLOT
                ):
                    self.add_plot_code(
                        question=item.item_name, plot_code=item.item_value
                    )

    def remake_collection(self, collection_name: str) -> bool:
        """
        Remakes a specified collection in the vector store.

        Args:
            collection_name (str): The name of the collection to be remade.

        Returns:
            bool: True if the collection was found and remade, False otherwise.

        Procedure:
            - Deletes the existing collection with the given name.
            - Recreates it using the embedding function.
        """
        for collection in vector_store_collections:
            if collection_name == collection:
                self.chroma_client.delete_collection(name=collection)
                setattr(
                    self,
                    f"{collection}_collection",
                    (
                        self.chroma_client.get_or_create_collection(
                            name=collection, embedding_function=self.embedding_function
                        )
                    ),
                )
                return True
        return False

    def _extract_documents(self, query_results) -> list:
        """
        Extracts and processes documents from the query results.

        Args:
            query_results (dict): The results obtained from a query, which may contain a "documents" key.

        Returns:
            list: A list of documents extracted from the query results.
                - If the "documents" key is not present or the query results are None, an empty list is returned.
                - If the documents are in JSON format, they are parsed and returned as a list of dictionaries.
        """
        if query_results is None:
            return []

        if "documents" in query_results:
            documents = query_results["documents"]
            if len(documents) == 1 and isinstance(documents[0], list):
                try:
                    documents = [json.loads(doc) for doc in documents[0]]
                except Exception:
                    return documents[0]

            return documents

    def get_related_ddl(self, question: str) -> list:
        """Retrieve related DDL statements based on the question."""
        return self._extract_documents(
            self.ddl_collection.query(
                query_texts=[question],
            )
        )

    def get_related_documentation(self, question: str) -> list:
        """Retrieve related documentation based on the question."""
        return self._extract_documents(
            self.documentation_collection.query(
                query_texts=[question],
            )
        )

    def get_related_sql_queries(self, user_input: str) -> list:
        """Retrieve similar question - SQL query pairs based on user input."""
        return self._extract_documents(
            self.sql_collection.query(
                query_texts=[user_input],
            )
        )

    def get_related_python_code(self, question: str) -> list:
        """Retrieve similar question - Python code pairs based on the question."""
        return self._extract_documents(
            self.python_collection.query(
                query_texts=[question],
            )
        )

    def get_related_plotting_code(self, question: str) -> list:
        """Retrieve similar question - plotting code pairs based on the question."""
        return self._extract_documents(
            self.plot_collection.query(
                query_texts=[question],
            )
        )

    def generate_embedding(self, data: str) -> list[float]:
        """Generate an embedding for the given data."""
        embedding = self.embedding_function([data])
        if len(embedding) == 1:
            return embedding[0]
        return embedding

    def add_question_sql(self, question: str, sql: str) -> str:
        """Add a question and its corresponding SQL query to the collection."""
        question_sql_json = json.dumps(
            {
                "question": question,
                "sql": sql,
            }
        )
        sql_id = ChromaDBVectorStore.deterministic_uuid(question_sql_json) + "-sql"
        self.sql_collection.add(
            documents=question_sql_json,
            embeddings=self.generate_embedding(question_sql_json),
            ids=sql_id,
        )
        return sql_id

    def add_ddl(self, ddl: str) -> str:
        """Add a DDL statement to the collection."""
        ddl_id = ChromaDBVectorStore.deterministic_uuid(ddl) + "-ddl"
        self.ddl_collection.add(
            documents=ddl,
            embeddings=self.generate_embedding(ddl),
            ids=ddl_id,
        )
        return ddl_id

    def add_documentation(self, documentation: str) -> str:
        """Add documentation to the collection."""
        doc_id = ChromaDBVectorStore.deterministic_uuid(documentation) + "-doc"
        self.documentation_collection.add(
            documents=documentation,
            embeddings=self.generate_embedding(documentation),
            ids=doc_id,
        )
        return doc_id

    def add_python_code(self, question: str, python_code: str):
        """Add a question and its corresponding Python code to the collection."""
        question_analysis_steps = json.dumps(
            {
                "question": question,
                "python_code": python_code,
            }
        )
        python_id = (
            ChromaDBVectorStore.deterministic_uuid(question_analysis_steps) + "-python"
        )
        self.python_collection.add(
            documents=question_analysis_steps,
            embeddings=self.generate_embedding(question_analysis_steps),
            ids=python_id,
        )
        return python_id

    def add_plot_code(self, question: str, plot_code: str):
        """Add a question and its corresponding plotting code to the collection."""
        question_plot_steps = json.dumps(
            {
                "question": question,
                "plot_code": plot_code,
            }
        )
        code_id = ChromaDBVectorStore.deterministic_uuid(question_plot_steps) + "-code"
        self.plot_collection.add(
            documents=question_plot_steps,
            embeddings=self.generate_embedding(question_plot_steps),
            ids=code_id,
        )
        return code_id
